// ---------------------------------------------------------------------------------
// Quick Battle 2 editor
//
// by Uwe Brinkschulte
//
// Handle Configuration File
// ----------------------------------------------------------------------------------

// ------------------------------ includes ---------------------------------------------

#include <afxwin.h>
#include "resrc1.h"
#include "qb2config.h"
#include "qb2edit.h"
#include "qb2text.h"

// ------------------------------ defines ----------------------------------------------

// ------------------------------ variables --------------------------------------------

// ------------------------------ external variables -----------------------------------

// path to KA main folder
extern CString   strPathToKA;

// ------------------------------  functions -------------------------------------------

//-----------------------------------
// test if a number (atoi is too lax, says ok if the first part is a number and the rest not)
//-----------------------------------
BOOL IsNumber(char *src)
{
  unsigned i;
  BOOL     bIsDecimal = FALSE;

  // check the string
  if (src[0]=='\0') return FALSE;
  for (i = 0; src[i]!='\0'; i++)
  {
    if ((src[i] < '0') || (src[i] > '9'))
    {
      // not a digit, we allow one decimal point
      if (src[i] == '.')
      {
        if (!bIsDecimal) bIsDecimal = TRUE;
                    else return FALSE;
      } else
      {
        // not a digit and not a decimal point.
        // In the first position we allow a - as well
        if ((i != 0) || (src[i] != '-')) return FALSE;
      }
    }
  }
  return TRUE;
}

//-----------------------------------
// get a substring
//-----------------------------------
void GetSubString(char *dest, const char *src, unsigned iStart, unsigned iEnd)
{
  unsigned i;

  // check start index (handles empty strings as well)
  if (iStart >= strlen(src))
  {
    // invalid
    dest[0] = '\0';
    return;
  }

  // check end index (strlen is greater 0 here, otherwise the start index check would have failed)
  if (iEnd >= strlen(src))
  {
    // limit end index
    iEnd = strlen(src) - 1;
  }

  // do it
  for (i = iStart; i <= iEnd; i++)
    dest[i - iStart] = src[i];
  dest[i - iStart] = '\0';
}

//-----------------------------------
// get the nth word of a string
//-----------------------------------
void GetNthWordInString(char *dest, char *src, unsigned n)
{
  unsigned i, iStart, nWord;

  i      = 0;
  iStart = 0;
  nWord  = 0;

  while (i < strlen(src))
  {
    // skip blanks and tabs
    for (; i < strlen(src); i++)
    {
      if ((src[i] != ' ') && (src[i] != 0x9)) break;
    }

    // a word found ?
    if (i < strlen(src))
    {
      // yes
      iStart = i;
      // find end of word
      for (; i < strlen(src); i++)
      {
        if ((src[i] == ' ') || (src[i] == 0x9)) break;
      }

      // is it the word we are lookin for ?
      nWord ++;
      if (nWord == n) break;
    }
  }

  // we have found the word ?
  if (nWord == n)
  {
    if (i>0) GetSubString(dest,src,iStart,i-1);
        else dest[0] = '\0';
  } else
  {
    dest[0] = '\0';
  }
}

//-----------------------------------
// tokenize a string
//-----------------------------------
void Tokenize(char *dest, char *src, char tokendelimiter)
{
  unsigned i;

  for (i = 0; i < strlen(src); i++)
  {
    // look for the token delimiter
    if (src[i] == tokendelimiter) break;
  }
  // copy the token
  if (i>0) GetSubString(dest, src, 0, i-1);
     else dest[0] = '\0';
}

//-----------------------------------
// remove all leading and tailing blanks from a string
//-----------------------------------
void StripString(char *dest, char *src)
{
  unsigned iStart;
  unsigned iEnd;

  // leading blanks and tabs
  for (iStart = 0; iStart < strlen(src); iStart++)
  {
    if ((src[iStart] != ' ') && (src[iStart] != 0x9)) break;
  }

  // ending blanks and tabs
  for (iEnd = strlen(src) - 1; iEnd >= 0; iEnd--)
  {
    if ((src[iEnd] != ' ') && (src[iEnd] != 0x9)) break;
  }

  // remove it
  if (iStart > iEnd)
  {
    // empty result
    dest[0] = '\0';
  } else
  {
    // cut the result
    GetSubString(dest, src, iStart, iEnd);
  }

}



// ------------------------------  classes  --------------------------------------------


// ----------- the config file class ---------------------------------------

//-----------------------------------
// the constructor
//-----------------------------------
CCfgFile::CCfgFile()
{
  pConfigFile = NULL;
}

//-----------------------------------
// open the file
//-----------------------------------
BOOL CCfgFile::Open(const char *fname)
{
  pConfigFile = fopen(fname,"r+t");
  if (pConfigFile == NULL) return FALSE;
  strncpy(strFName,fname,PATH_LENGTH);
  return TRUE;
}

//-----------------------------------
// create the file
//-----------------------------------
BOOL CCfgFile::Create(const char *fname)
{
  pConfigFile = fopen(fname,"w+t");
  if (pConfigFile == NULL) return FALSE;
  strncpy(strFName,fname,PATH_LENGTH);
  return TRUE;
}

//-----------------------------------
// read a line
//-----------------------------------
BOOL CCfgFile::ReadLine(char *line)
{
  char c;
  int  i = 0;
  BOOL gotAtLeastOne = FALSE;

  // file opened ?
  if (pConfigFile == NULL) return FALSE;

  // read a line
  while (fread(&c,1,1,pConfigFile) == 1)
  {
    gotAtLeastOne = TRUE;
    if ((c != '\n') && (i < LINE_LENGTH - 1))
    {
      line[i] = c;
      i++;
    }
    if (c == '\n') break;
  }
  // terminate string
  line[i] = '\0';

  return (gotAtLeastOne);
}

//-----------------------------------
// write a line
//-----------------------------------
BOOL CCfgFile::WriteLine(char *line)
{
  // file opened ?
  if (pConfigFile == NULL) return FALSE;

  // write the line
  return (fprintf(pConfigFile, "%s\n", line) != EOF);

}

//-----------------------------------
// reset the file
//-----------------------------------
BOOL CCfgFile::Reset()
{
  // file opened ?
  if (pConfigFile == NULL) return FALSE;

  // close the file
  fclose(pConfigFile);

  // open it to override old contents
  pConfigFile = fopen(strFName,"wt");
  
  return (pConfigFile != NULL);
}

//-----------------------------------
// close the file
//-----------------------------------
void CCfgFile::Close()
{
  if (pConfigFile != NULL) fclose(pConfigFile);
}


// ----------  the ship data class, contains all info of a specific ship ------

//-----------------------------------
// the constructor
//-----------------------------------
CShipData::CShipData()
{
  // init all protected variables
  m_ShipClass = "";
  m_ShipName = "";

  // no data set yet
  m_ShipClassSet = FALSE;
  m_ShipNameSet = FALSE;
}

// ----------  the ship list class, contains all info of a list of ships ------

//-----------------------------------
// look if a ship exists
// if bDefault is FALSE, we search among all ships where the user has specified the name
// if bDefault is TRUE, we search among all ships where the user has not specified the name (the default name is used)
// (To construct the default shipname name the type ('F' = friend, 'E' = escort, 'O' = enemy, 
//  'H' = hulk, 'N' = neutral) and system number must be given.
//-----------------------------------
BOOL CShipList::ExistsShip(const char *strShipName, BOOL bDefault, char cType, unsigned uSysNum)
{
  POSITION  pos;
  CShipData *pShip;
  unsigned  i;
  char      defname[MAX_QB2_LINELENGTH + 1];

  // iterate through the list
  i = 0;
  for (pos = GetHeadPosition(); pos != NULL; )
  {
    i++;
    pShip = (CShipData *)GetNext(pos);
    // shall we check against the given names or the default names ?
    if (!bDefault)
    {
      // check the given names
      if (pShip->IsSetShipName())
      {
        // the ship has a name, check it
        if (strcmp(strShipName,pShip->GetShipName()) == 0) return TRUE;
      }
    } else
    {
      // check the default names
      if (!pShip->IsSetShipName())
      {
        // create the default shipname
        if (cType == 'E')
          // an escort
          sprintf(defname,"%s %c%u", pShip->GetShipClass(), cType, i);
        else
          // friend, enemy, neutral or hulk
          sprintf(defname,"%s %c%u.%u", pShip->GetShipClass(), cType, uSysNum, i);
        // check it
        if (strcmp(strShipName,defname) == 0) return TRUE;
      }
    }
  }
  return FALSE;
}

// ----------  the dialog data class, contains all info of a specific dialog ------

//-----------------------------------
// the constructor
//-----------------------------------
CDialogData::CDialogData()
{
  // init all protected variables
  m_Speaker = "";
  m_DialogLine = "";
  m_VoiceFile = "";

  // no data set yet
  m_SpeakerSet = FALSE;
  m_DialogLineSet = FALSE;
  m_VoiceFileSet = FALSE;
}


// ---------------  the title data class, contains all info of a title ------

//-----------------------------------
// the constructor
//-----------------------------------
CTitleData::CTitleData()
{
  // init all protected variables
  m_Title = "";
  // the list is guarantied to be empty by the constructor, so nothing has to be done here

  // no data set yet
  m_TitleSet = FALSE;
}


// ---------------  the epilog data class, contains all info of an epilog ------

//-----------------------------------
// the constructor
//-----------------------------------
CEpilogData::CEpilogData()
{
  // the list is guarantied to be empty by the constructor, so nothing has to be done here
}


// -----------  the general mode data class, contains all general mode info ------

//-----------------------------------
// the constructor
//-----------------------------------
CGeneralModeData::CGeneralModeData()
{
  // init all protected variables
  m_ShipChangeMode = NoShipChange;
  m_CollisionMode = CollisionsOn;
  m_WarpOutMode = Anytime;
  m_PerfectCloakMode = PerfectCloakOff;

  // no data set yet
  m_ShipChangeModeSet = FALSE;
  m_CollisionModeSet = FALSE;
  m_WarpOutModeSet = FALSE;
  m_PerfectCloakModeSet = FALSE;
}

// -----------  the jump option data class, contains all jump gate options ------

//-----------------------------------
// the constructor
//-----------------------------------
CJumpOptionsData::CJumpOptionsData()
{
  // init all protected variables
  m_ActDistance = 0;

  // no data set yet
  m_GateHiddenSet = FALSE;
  m_JumpOnlySet = FALSE;
  m_ClearJumpSet = FALSE;
  m_NoSpinSet = FALSE;
  m_ShowActivationSet = FALSE;
  m_ActOnHailSet = FALSE;
  m_ActDistanceSet = FALSE;
}

// -----------  the self destruct data class, contains all self destruct info ------

//-----------------------------------
// the constructor
//-----------------------------------
CSelfDestructData::CSelfDestructData()
{
  // init all protected variables
  m_Mode = StandardDestruct;
  m_Code = "";
  m_OverkillDistance = 0;

  // no data set yet
  m_ModeSet = FALSE;
}

// ----------- the system data class, contains all info of a specific solar system --------------

//-----------------------------------
// the constructor
//-----------------------------------
CSystemData::CSystemData()
{
  // init all protected variables
  m_SystemName = "";
  m_Stardate = 0;
  m_JumpTime = 0;
  m_JumpDestination = "";
  m_EnemyPosX = 0;
  m_EnemyPosY = 0;
  m_EnemyPosZ = 0;
  m_EnemyPosPlanet = "";
  m_FriendPosX = 0;
  m_FriendPosY = 0;
  m_FriendPosZ = 0;
  m_FriendPosPlanet = "";
  m_PlayerPosX = 0;
  m_PlayerPosY = 0;
  m_PlayerPosZ = 0;
  m_PlayerPosPlanet = "";
  m_NeutralPosX = 0;
  m_NeutralPosY = 0;
  m_NeutralPosZ = 0;
  m_NeutralPosPlanet = "";
  m_HulkPosX = 0;
  m_HulkPosY = 0;
  m_HulkPosZ = 0;
  m_HulkPosPlanet = "";
  m_JumpGatePosX = 0;
  m_JumpGatePosY = 0;
  m_JumpGatePosZ = 0;
  m_JumpGatePosPlanet = "";
  m_NeutralBehavior = EnemyWhenAttacked;
  m_NeutralBehaviorTime = 0;
  m_PlayerWarpDisable = FALSE;
  m_FriendWarpDisable = FALSE;
  m_EnemyWarpDisable = FALSE;
  m_NeutralWarpDisable = FALSE;
  m_BattleMode = AIBattle;
  m_FriendWarpInTime = 0;
  m_EnemyWarpInTime = 0;
  m_NeutralWarpInTime = 0;
  m_FriendWarpOutTime = 0;
  m_EnemyWarpOutTime = 0;
  m_NeutralWarpOutTime = 0;
  m_FriendJoinEscortTime = 0;

  // the lists are guarantied to be empty by the constructor, so nothing has to be done here

  // no data set yet
  m_SystemNameSet = FALSE;
  m_StardateSet = FALSE;
  m_JumpTimeSet = FALSE;
  m_JumpDestinationSet = FALSE;
  m_EnemyPosSet = FALSE;
  m_EnemyPosPlanetSet = FALSE;
  m_FriendPosSet = FALSE;
  m_FriendPosPlanetSet = FALSE;
  m_PlayerPosSet = FALSE;
  m_PlayerPosPlanetSet = FALSE;
  m_NeutralPosSet = FALSE;
  m_NeutralPosPlanetSet = FALSE;
  m_HulkPosSet = FALSE;
  m_HulkPosPlanetSet = FALSE;
  m_JumpGatePosSet = FALSE;
  m_JumpGatePosPlanetSet = FALSE;
  m_NeutralBehaviorSet = FALSE;
  m_NeutralBehaviorTimeSet = FALSE;
  m_PlayerWarpDisableSet = FALSE;
  m_FriendWarpDisableSet = FALSE;
  m_EnemyWarpDisableSet = FALSE;
  m_NeutralWarpDisableSet = FALSE;
  m_BattleModeSet = FALSE;
  m_FriendWarpInTimeSet = FALSE;
  m_EnemyWarpInTimeSet = FALSE;
  m_NeutralWarpInTimeSet = FALSE;
  m_FriendWarpOutTimeSet = FALSE;
  m_EnemyWarpOutTimeSet = FALSE;
  m_NeutralWarpOutTimeSet = FALSE;
  m_FriendJoinEscortTimeSet = FALSE;
}

// ----------  the system list class, contains all info of all systems ------

//-----------------------------------
// look if a system exists
//-----------------------------------
BOOL CSystemList::ExistsSystem(const char *strSystemName)
{
  POSITION     pos;
  CSystemData *pSystem;
  CString      sName;

  // iterate through the list
  for (pos = GetHeadPosition(); pos != NULL; )
  {
    pSystem = (CSystemData *)GetNext(pos);
    // check the name
    if (pSystem->IsSetSystemName())
    {
      // the system has a name, check it (not case sensitive)
      sName = pSystem->GetSystemName();
      if (sName.CompareNoCase(strSystemName) == 0) return TRUE;
    }
  }
  return FALSE;
}


// ----------- the main configuration class -----------------------------------

//-----------------------------------
// the constructor
//-----------------------------------
CConfiguration::CConfiguration()
{
  // init some stuff
  strConfigName = "";
  uParsedLine = 0;
  pCurrentSystem = NULL;
  bSampleRemarks = TRUE;
}

//-----------------------------------
// the destructor
//-----------------------------------
CConfiguration::~CConfiguration()
{
}

//-----------------------------------
// remove remarks
//-----------------------------------
void CConfiguration::RemoveRemarks(char *dest, char *src)
{

  int   i, iRemLen;
  char  str[10];
  BOOL  RemarkFound;

  RemarkFound = FALSE;
  iRemLen = strlen(REMARK) - 1;
  
  // look for a remark
  for (i = 0; i < ((int)strlen(src) - iRemLen); i++)
  {
    GetSubString(str, src, i, i + iRemLen);
    if (strcmp(str,REMARK) == 0)
    {
      RemarkFound = TRUE;
      break;
    }
  }

  // remark found ?
  if (RemarkFound)
  {
    // remark found, remove it
    if (i == 0)
    {
      // whole line is a remark
      dest[0] = '\0';
    } else
    {
      // cut the string
      GetSubString(dest,src,0,i-1);
    }
  } else
  {
    // no remark found, copy whole line
    strcpy(dest,src);
  }
}

//-----------------------------------
// sample remarks
// (samples remarks up to the first keyword or empty line, checks the qb2edit brand)
//-----------------------------------
BOOL CConfiguration::SampleRemarks(char *src)
{
  char           str[MAX_QB2_LINELENGTH + 1];
  char           str1[10];
  CMessageDialog Message;
  CString        *pRemark, *pBrand;

  // check if the first remark is the qb2edit brand
  if (Remarks.GetCount()==0)
  {
    if (strcmp(src, QB2EDIT_BRAND) != 0)
    {
      // no, give a warning that this file was not created by qb2edit
      Message.Report(WARN_BRAND, WARN_HEADER);
      // add the brand
      pBrand = new CString;
      if (pBrand == NULL)
      {
        ReportConfigError(ERR_OUTOFMEMORY, FALSE);
        return FALSE;
      }
      // add the brand to the list (first remove the remark characters)
      GetSubString(str, QB2EDIT_BRAND, strlen(REMARK), strlen(QB2EDIT_BRAND) - 1);
      *pBrand = str;
      Remarks.AddTail((CObject *)pBrand);
    }
  }


  // are we still sampling remarks ?
  if (bSampleRemarks)
  {
    // first strip the string
    StripString(str, src);
    // check if it is remark line
    GetSubString(str1,str,0,strlen(REMARK)-1);
    if (strcmp(str1,REMARK) == 0)
    {
      // yes it is, sample it
      pRemark = new CString;
      if (pRemark == NULL)
      {
        ReportConfigError(ERR_OUTOFMEMORY, FALSE);
        return FALSE;
      }
      // remove remark characters
      GetSubString(str, str, strlen(REMARK), strlen(str)-1);
      *pRemark = str;
      // add it to the list
      Remarks.AddTail((CObject *)pRemark);
    } else
    {
      // no, stop sampling remarks
      bSampleRemarks = FALSE;
    }
  }
  return TRUE;
}

//-----------------------------------
// report a configuration error
//-----------------------------------
void CConfiguration::ReportConfigError(const char *ErrorMessage, BOOL printLine)
{
  CString Header, Message;
  CErrorDialog ErrorDialog(NULL);
 
  // create the header
  Header = ERR_HEADER1 + strConfigName + ".qb2";
  // create the message
  Message = ErrorMessage;
  // print line number ?
  if (printLine)
  {
    // add line number to message
    char str[20];
    sprintf(str,"%u",uParsedLine);
    Message = Message + str;
  }
  ErrorDialog.Report(Message, Header);
}

//-----------------------------------
// do some basic parameter test
//-----------------------------------
BOOL CConfiguration::CheckParam(char *strParam)
{

  // check the parameter
  if (strParam[0] == '\0')
  {
    ReportConfigError(ERR_EMPTYPARAM, TRUE);
    return FALSE;
  }

  if ((strParam[0] == '"') || (strParam[strlen(strParam)-1] == '"'))
  {
    ReportConfigError(ERR_INVALIDQUOTES, TRUE);
    return FALSE;
  }

  if (strParam[0] == '=')
  {
    ReportConfigError(ERR_INVALIDEQUAL, TRUE);
    return FALSE;
  }

  return TRUE;
}

//-----------------------------------
// check if a shipname is unique
// if bCheckDefault is false it checks against the ship names given by the user
// if bCheckDefault is true it checks against the default ship names (the ships where the user hasn't given a name)
// if bReportError is true, an error message is generated, otherwise just false is returned
// if bShowLine is true the errorness line number of the configfile is included in the error message
// uWhere gives back the system number (0..6), where the second occurence of the name was found
// (a 0 is returned when the ship was the player or the escort (not assigned to a system)
// uGroup gives back the group of this ship (0 = player, 1 = escort, 2 = friend, 3 = enemy, 4 = neutral, 5 = hulk, 6 = jump gate)
//-----------------------------------
BOOL CConfiguration::CheckShipName(const char *strShipName, BOOL bCheckDefault, BOOL bReportError, BOOL bShowLine,
                                   unsigned *uWhere, unsigned *uGroup)
{
  CSystemData *pSystem;
  CShipList   *pShipList;
  POSITION    pos;
  unsigned    iSys;
  CString     strErrorMessage;
  char        defname[MAX_QB2_LINELENGTH + 1];

  // look if the shipname is unique

  // setup the error message
  strErrorMessage = ERR_SHIPNAME;
  strErrorMessage = strErrorMessage + strShipName;

  // check against name of playership
  if (!bCheckDefault)
  {
    // check against the given name
    if (PlayerShip.IsSetShipName())
    { 
      if (strcmp(strShipName,PlayerShip.GetShipName()) == 0)
      {
        if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
        if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
        *uWhere = 0;
        *uGroup = 0;
        return FALSE;
      }
    }
  } else
  {
    // check against the default name
    if (!PlayerShip.IsSetShipName())
    { 
      if (strcmp(strShipName,PlayerShip.GetShipClass() + " P") == 0)
      {
        if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
        if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
        *uWhere = 0;
        *uGroup = 0;
        return FALSE;
      }
    }
  }

  // check against name of the escorts
  if (Escort.ExistsShip(strShipName, bCheckDefault, 'E', 0))
  {
    if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
    if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
    *uWhere = 0;
    *uGroup = 1;
    return FALSE;
  }
  // check against all friends, enemies, neutrals and hulks in all systems
  iSys = 0;
  for (pos = Systems.GetHeadPosition(); pos != NULL; )
  {

    // iterate through systems
    iSys++;
    pSystem = (CSystemData *)Systems.GetNext(pos);

    // check the friends
    pShipList = pSystem->GetFriends();
    if (pShipList->ExistsShip(strShipName, bCheckDefault, 'F', iSys))
    {
      if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
      if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
      *uWhere = iSys;
      *uGroup = 2;
      return FALSE;
    }

    // check the enemies
    pShipList = pSystem->GetEnemies();
    if (pShipList->ExistsShip(strShipName, bCheckDefault, 'O', iSys))
    {
      if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
      if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
      *uWhere = iSys;
      *uGroup = 3;
      return FALSE;
    }

    // check the neutrals
    pShipList = pSystem->GetNeutrals();
    if (pShipList->ExistsShip(strShipName, bCheckDefault, 'N', iSys))
    {
      if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
      if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
      *uWhere = iSys;
      *uGroup = 4;
      return FALSE;
    }

    // check the hulks
    pShipList = pSystem->GetHulks();
    if (pShipList->ExistsShip(strShipName, bCheckDefault, 'H', iSys))
    {
      if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
      if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
      *uWhere = iSys;
      *uGroup = 5;
      return FALSE;
    }

    // check the jump gate
    if (!bCheckDefault)
    {
      // check against the given name
      if (pSystem->GetJumpGate()->IsSetShipName())
      { 
        if (strcmp(strShipName,pSystem->GetJumpGate()->GetShipName()) == 0)
        {
          if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
          if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
          *uWhere = iSys;
          *uGroup = 6;
          return FALSE;
        }
      }
    } else
    {
      // check against the default name
      if (!pSystem->GetJumpGate()->IsSetShipName())
      { 
        sprintf(defname,"Jump Gate %u", iSys);
        if (strcmp(strShipName,defname) == 0)
        {
          if (bShowLine) strErrorMessage = strErrorMessage + ERR_SHIPNAME1;
          if (bReportError) ReportConfigError(strErrorMessage, bShowLine);
          *uWhere = iSys;
          *uGroup = 6;
          return FALSE;
        }
      }
    }
  }

  return TRUE;
}

//-----------------------------------
// check all user given ship names against all default ship names
// uWhereGiven/uWhereDefault gives back the system numbers (0..6) of the conflict
// (a 0 is returned when the ship was the player or the escort (not assigned to a system)
// uGroupGiven/uGroupDefault gives back the groups of the conflict (0 = player, 1 = escort, 2 = friend, 3 = enemy, 4 = neutral, 5 = hulk, 6 = jump gate)
// strShipName gives the conflicting name
//-----------------------------------
BOOL CConfiguration::CheckAllGivenAgainstDefaultShipNames(unsigned *uWhereGiven, unsigned *uGroupGiven,
                                              unsigned *uWhereDefault, unsigned *uGroupDefault, CString *strShipName)
{
  CSystemData *pSystem;
  CShipList   *pShipList;
  CShipData   *pShip;
  POSITION    posship, possystem;
  unsigned    iSys;
  CString     strErrorMessage;

  // first we check the name of the playership
  if (PlayerShip.IsSetShipName())
  {
    // against all default names
    if (!CheckShipName(PlayerShip.GetShipName(), TRUE, FALSE, FALSE, uWhereDefault, uGroupDefault))
    {
      *uWhereGiven = 0;
      *uGroupGiven = 0;
      *strShipName = PlayerShip.GetShipName();
      return FALSE;
    }
  }

  // now lets check all escorts
  for (posship = Escort.GetHeadPosition(); posship != NULL; )
  {
    pShip = (CShipData *)Escort.GetNext(posship);
    if (pShip->IsSetShipName())
    {
      // against all default names
      if (!CheckShipName(pShip->GetShipName(), TRUE, FALSE, FALSE, uWhereDefault, uGroupDefault))
      {
        *uWhereGiven = 0;
        *uGroupGiven = 1;
        *strShipName = pShip->GetShipName();
        return FALSE;
      }
    }
  }

  // now lets check all friends, enemies, neutrals and hulks in all systems
  iSys = 0;
  for (possystem = Systems.GetHeadPosition(); possystem != NULL; )
  {
    // iterate through systems
    iSys++;
    pSystem = (CSystemData *)Systems.GetNext(possystem);

    // check the friends
    pShipList = pSystem->GetFriends();
    for (posship = pShipList->GetHeadPosition(); posship != NULL; )
    {
      pShip = (CShipData *)pShipList->GetNext(posship);
      if (pShip->IsSetShipName())
      {
        // against all default names
        if (!CheckShipName(pShip->GetShipName(), TRUE, FALSE, FALSE, uWhereDefault, uGroupDefault))
        {
          *uWhereGiven = iSys;
          *uGroupGiven = 2;
          *strShipName = pShip->GetShipName();
          return FALSE;
        }
      }
    }

    // check the enemies
    pShipList = pSystem->GetEnemies();
    for (posship = pShipList->GetHeadPosition(); posship != NULL; )
    {
      pShip = (CShipData *)pShipList->GetNext(posship);
      if (pShip->IsSetShipName())
      {
        // against all default names
        if (!CheckShipName(pShip->GetShipName(), TRUE, FALSE, FALSE, uWhereDefault, uGroupDefault))
        {
          *uWhereGiven = iSys;
          *uGroupGiven = 3;
          *strShipName = pShip->GetShipName();
          return FALSE;
        }
      }
    }

    // check the neutrals
    pShipList = pSystem->GetNeutrals();
    for (posship = pShipList->GetHeadPosition(); posship != NULL; )
    {
      pShip = (CShipData *)pShipList->GetNext(posship);
      if (pShip->IsSetShipName())
      {
        // against all default names
        if (!CheckShipName(pShip->GetShipName(), TRUE, FALSE, FALSE, uWhereDefault, uGroupDefault))
        {
          *uWhereGiven = iSys;
          *uGroupGiven = 4;
          *strShipName = pShip->GetShipName();
          return FALSE;
        }
      }
    }

    // check the hulks
    pShipList = pSystem->GetHulks();
    for (posship = pShipList->GetHeadPosition(); posship != NULL; )
    {
      pShip = (CShipData *)pShipList->GetNext(posship);
      if (pShip->IsSetShipName())
      {
        // against all default names
        if (!CheckShipName(pShip->GetShipName(), TRUE, FALSE, FALSE, uWhereDefault, uGroupDefault))
        {
          *uWhereGiven = iSys;
          *uGroupGiven = 5;
          *strShipName = pShip->GetShipName();
          return FALSE;
        }
      }
    }

    // check the jump gate
    if (pSystem->GetJumpGate()->IsSetShipName())
    {
      // against all default names
      if (!CheckShipName(pSystem->GetJumpGate()->GetShipName(), TRUE, FALSE, FALSE, uWhereDefault, uGroupDefault))
      {
        *uWhereGiven = iSys;
        *uGroupGiven = 6;
        *strShipName = pSystem->GetJumpGate()->GetShipName();
        return FALSE;
      }
    }

  }

  // all is fine
  return TRUE;
}


//-----------------------------------
// check some critical parameters for stat battle mode
//-----------------------------------
BOOL CConfiguration::CheckStatBattleMode(CSystemData *pSystem, BOOL bReportError)
{
  unsigned enemynum, friendnum;
  CString  strErrorMessage;

  if (pSystem != NULL)
  {
    if (pSystem->IsSetBattleMode() && (pSystem->GetBattleMode() == StatisticBattle))
    {
      // we have statistic battle mode defined, do some checking

      // only one of enemy position, friend position or neutral position is allowed using stat battle mode
      if (((pSystem->IsSetEnemyPos() || pSystem->IsSetEnemyPosPlanet()) && (pSystem->IsSetFriendPos() || pSystem->IsSetFriendPosPlanet())) ||
          ((pSystem->IsSetEnemyPos() || pSystem->IsSetEnemyPosPlanet()) && (pSystem->IsSetNeutralPos() || pSystem->IsSetNeutralPosPlanet())) ||
          ((pSystem->IsSetNeutralPos() || pSystem->IsSetNeutralPosPlanet()) && (pSystem->IsSetFriendPos() || pSystem->IsSetFriendPosPlanet())))
      {
        strErrorMessage = ERR_STATPOS;
        if (bReportError) ReportConfigError(strErrorMessage + pSystem->GetSystemName(), FALSE);
        return FALSE;
      }

      // at least one system resident enemy and friend required using stat battle mode
      enemynum =  (pSystem->GetEnemies())->GetCount();
      friendnum = (pSystem->GetFriends())->GetCount();
      if ((enemynum == 0) || (friendnum == 0))
      {
        strErrorMessage = ERR_STATRESSHIP;
        if (bReportError) ReportConfigError(strErrorMessage + pSystem->GetSystemName(), FALSE);
        return FALSE;
      }

      // friends warp in later is not allowed using stat battle mode (otherwise the enemies may be alone and do nothing in the system)
      if (pSystem->IsSetFriendWarpInTime())
      {
        strErrorMessage = ERR_STATNOWARPIN;
        if (bReportError) ReportConfigError(strErrorMessage + pSystem->GetSystemName(), FALSE);
        return FALSE;
      }

      // friends becoming escorts is not allowed using stat battle mode
      if (pSystem->IsSetFriendJoinEscortTime())
      {
        strErrorMessage = ERR_STATNOJOINESCORT;
        if (bReportError) ReportConfigError(strErrorMessage + pSystem->GetSystemName(), FALSE);
        return FALSE;
      }
    }
  }
  return TRUE;

}

//-----------------------------------
// check jump destination systems
//-----------------------------------
BOOL CConfiguration::CheckJumpDestination(CString DestSystemName)
{
  POSITION    pos;
  CSystemData *pSystem;
  BOOL        found;

  // Checks if the given system name is contained in the battle and therefor a valid jump destination
  // Loop through all systems to check
  found = FALSE;
  for (pos = Systems.GetHeadPosition(); pos != NULL; )
  {
    pSystem = (CSystemData *)Systems.GetNext(pos);
    // compare the destination with the system names
    if (DestSystemName.CompareNoCase(pSystem->GetSystemName()) == 0)
    {
      // we have found it, it is ok
      found = TRUE;
      break;
    }
  }
  // we have not found the destination system ?
  // if not this is an error!
  return found;
}

//-----------------------------------
// check some other critical parameters for a system
//-----------------------------------
BOOL CConfiguration::CheckOthers(CSystemData *pSystem)
{
  CString  strErrorMessage;

  // join escort cannot be defined together with warp out for friends
  if (pSystem->IsSetFriendJoinEscortTime() && pSystem->IsSetFriendWarpOutTime())
  {
     strErrorMessage = ERR_JOINWARPOUT;
     ReportConfigError(strErrorMessage + pSystem->GetSystemName(), FALSE);
     return FALSE;
  }

  // player defined a jump gate but no position and no time
  if (pSystem->GetJumpGate()->IsSetShipClass() && !(pSystem->IsSetJumpGatePos()) &&
      !(pSystem->IsSetJumpGatePosPlanet()) && !(pSystem->IsSetJumpTime()))
  {
     strErrorMessage = ERR_JUMPGATENOPOS;
     ReportConfigError(strErrorMessage + pSystem->GetSystemName(), FALSE);
     return FALSE;
  }

  // player defined a jump gate but no destination
  if (pSystem->GetJumpGate()->IsSetShipClass() &&
      !(pSystem->IsSetJumpDestination()))
  {
     strErrorMessage = ERR_JUMPGATENODEST;
     ReportConfigError(strErrorMessage + pSystem->GetSystemName(), FALSE);
     return FALSE;
  }

  return TRUE;
}

//-----------------------------------
// a final configuration check
//-----------------------------------
BOOL CConfiguration::FinalConfigCheck(BOOL ExtendedErrorMessage)
{
  CSystemData *pSystem;
  CString     strErrorMessage;
  POSITION    pos;

  // is there at least one system defined ?
  if (Systems.GetCount() == 0)
  {
    // no, this is an error
    ReportConfigError(ERR_NOSYSDEFINED, FALSE);
    return FALSE;
  }

  // is there a player ship defined ?
  if (!PlayerShip.IsSetShipClass())
  {
    // no, this is an error
    ReportConfigError(ERR_NOPLAYERDEF, FALSE);
    return FALSE;
  }

  // are all the jump destination systems defined ?
  for (pos = Systems.GetHeadPosition(); pos != NULL; )
  {
    // iterate through systems
    pSystem = (CSystemData *)Systems.GetNext(pos);
    // we have a jump destination ?
    if (pSystem->IsSetJumpDestination())
    {
      // check if valid
      if (!CheckJumpDestination(pSystem->GetJumpDestination()))
      {
        // standard message ?
        if (!ExtendedErrorMessage)
        {
          strErrorMessage = ERR_UNKNOWNJUMPDEST;
          ReportConfigError(strErrorMessage + pSystem->GetJumpDestination(), FALSE);
        } else
        // extended message
        {
          strErrorMessage = ERR_JUMPDESTNOTDEFINED1;
          strErrorMessage = strErrorMessage + pSystem->GetJumpDestination();
          strErrorMessage = strErrorMessage + ERR_JUMPDESTNOTDEFINED2;
          strErrorMessage = strErrorMessage + pSystem->GetSystemName();
          CMessageDialog Message;
          Message.Report(strErrorMessage + ".", ERR_HEADER);
        }
        return FALSE;
      }
    }
  }
  return TRUE;
}


//-----------------------------------
// setup a new configuration (only needed when not parsed)
//-----------------------------------
BOOL CConfiguration::Setup()
{
  char           str[MAX_QB2_LINELENGTH + 1];
  CString        *pBrand;
  CErrorDialog   ErrorDialog;

  // setting up a new configuration is simple: everything has to be empty
  // except the remarks. The first remark has to be the edit brand

  pBrand = new CString;
  if (pBrand == NULL)
  {
    ErrorDialog.Report(ERR_OUTOFMEMORY, ERR_HEADER);
    return FALSE;
  }
  // add the brand to the list (first remove the remark characters)
  GetSubString(str, QB2EDIT_BRAND, strlen(REMARK), strlen(QB2EDIT_BRAND) - 1);
  *pBrand = str;
  Remarks.AddTail((CObject *)pBrand);

  return TRUE;
}


//-----------------------------------
// parse a existing configuration
// if Path is NULL the default path to the ka scripting folder is used
// CheckOnly is true means only checking the configuration, no editing.
// (some messages and remark sampling is supressed in this mode) 
//-----------------------------------
BOOL CConfiguration::Parse(const char *ConfigName, const char *Path, BOOL CheckOnly)
{
  CString  strConfigFileName;
  BOOL     bStopParsing;
  char     line[LINE_LENGTH];
  char     line1[MAX_QB2_LINELENGTH + 1];
  char     strKeyword[MAX_QB2_LINELENGTH + 1];
  char     strParam[MAX_QB2_LINELENGTH + 1];

  // get the name and path of the config file
  if (Path == NULL) 
  {
    strConfigFileName = strPathToKA + QB2_PATH + ConfigName + ".qb2";
  } else
  {
    strConfigFileName = Path;
    strConfigFileName = strConfigFileName + ConfigName + ".qb2";
  }
  strConfigName = ConfigName;
  uParsedLine = 0;

  // open the config file
  if (!ConfigFile.Open(strConfigFileName))
  {
    // an unexpected error
    ReportConfigError(ERR_CFGOPEN, FALSE);
    return FALSE;
  }

  // check the config file
  if (!ConfigFile.ReadLine(line))
  {
    // empty file is invalid
    ReportConfigError(ERR_CFGINVALID, FALSE);
    ConfigFile.Close();
    return FALSE;
  }
  if (strcmp(line,QB2_MAGIC) != 0)
  {
    // not the proper magic
    ReportConfigError(ERR_CFGINVALID, FALSE);
    ConfigFile.Close();
    return FALSE;
  }
  if (!ConfigFile.ReadLine(line))
  {
    // no version line
    ReportConfigError(ERR_CFGINVALID, FALSE);
    ConfigFile.Close();
    return FALSE;
  }
  if ((strcmp(line,QB2_VERSION1) != 0) && (strcmp(line,QB2_VERSION2) != 0) && 
      (strcmp(line,QB2_VERSION3) != 0) && (strcmp(line,QB2_VERSION4) != 0))
  {
    // not the proper version
    ReportConfigError(ERR_CFGVERSION, FALSE);
    ConfigFile.Close();
    return FALSE;
  }

  // start with line 3
  uParsedLine = 3;

  // parse the file to its end
  bStopParsing = FALSE;
  while (ConfigFile.ReadLine(line))
  {

    // line to long ?
    if (strlen(line) > MAX_QB2_LINELENGTH)
    {
      ReportConfigError(ERR_LINETOOLONG, TRUE);
      ConfigFile.Close();
      return FALSE;
    }

    // sample the initial remarks (not in checkonly mode, because we don't need it there and we don't 
    // need some warnings produced by this routine)
    if (!CheckOnly)
    {
      if (!SampleRemarks(line)) return FALSE;
    }

    // remove remarks for further processing
    RemoveRemarks(line, line);

    // try to find a keyword
    Tokenize(line1, line, '=');

    // do we have a keyword ?
    if (strlen(line1) != strlen(line))
    {
      // yes, isolate keyword (remove leading and ending blanks)
      StripString(strKeyword, line1);
      // isolate the parameter
      GetSubString(strParam, line, strlen(line1)+1,strlen(line)-1);
    } else
    {
      // no keyword, the whole line is a parameter
      strKeyword[0] = '\0';
      strcpy(strParam, line);
    }

    // remove leading and ending blanks or tabs from the parameter
    StripString(strParam,strParam);


    // examine the keyword

    // the player ?
    if (strcmp(strKeyword, KEY_PLAYER) == 0)
    {
      bStopParsing = DefinePlayer(strParam);
    } else

    // player's escort ?
    if (strcmp(strKeyword, KEY_ESCORT) == 0)
    {
      bStopParsing = DefineEscort(strParam);
    } else

    // a friend ?
    if (strcmp(strKeyword, KEY_FRIEND) == 0)
    {
      bStopParsing = DefineFriend(strParam);
    } else

    // an enemy ?
    if (strcmp(strKeyword, KEY_ENEMY) == 0)
    {
      bStopParsing = DefineEnemy(strParam);
    } else

    // a neutral ?
    if (strcmp(strKeyword, KEY_NEUTRAL) == 0)
    {
      bStopParsing = DefineNeutral(strParam);
    } else

    // a hulk ?
    if (strcmp(strKeyword, KEY_HULK) == 0)
    {
      bStopParsing = DefineHulk(strParam);
    } else

    // a jump gate ?
    if (strcmp(strKeyword, KEY_JUMPGATE) == 0)
    {
      bStopParsing = DefineJumpGate(strParam);
    } else

    // a system ?
    if (strcmp(strKeyword, KEY_SYSTEM) == 0)
    {
       bStopParsing = DefineSystem(strParam);
    } else
 
    // player's position in a system ?
    if (strcmp(strKeyword, KEY_PLAYERPOS) == 0)
    {
       bStopParsing = DefinePlayerPosition(strParam);
    } else

    // friends's position in a system ?
    if (strcmp(strKeyword, KEY_FRIENDPOS) == 0)
    {
       bStopParsing = DefineFriendPosition(strParam);
    } else

    // enemy's position in a system ?
    if (strcmp(strKeyword, KEY_ENEMYPOS) == 0)
    {
       bStopParsing = DefineEnemyPosition(strParam);
    } else

    // neutral's position in a system ?
    if (strcmp(strKeyword, KEY_NEUTRALPOS) == 0)
    {
       bStopParsing = DefineNeutralPosition(strParam);
    } else

    // hulk's position in a system ?
    if (strcmp(strKeyword, KEY_HULKPOS) == 0)
    {
       bStopParsing = DefineHulkPosition(strParam);
    } else

    // jump gate position in a system ?
    if (strcmp(strKeyword, KEY_JUMPPOS) == 0)
    {
       bStopParsing = DefineJumpGatePosition(strParam);
    } else

    // jump time in a system ?
    if (strcmp(strKeyword, KEY_JUMPTIME) == 0)
    {
       bStopParsing = DefineJumpTime(strParam);
    } else

    // jump destination in a system ?
    if (strcmp(strKeyword, KEY_JUMPDEST) == 0)
    {
       bStopParsing = DefineJumpDestination(strParam);
    } else

    // jump options in a system ?
    if (strcmp(strKeyword, KEY_JUMPOPTION) == 0)
    {
       bStopParsing = DefineJumpOptions(strParam);
    } else

    // a stardate ?
    if (strcmp(strKeyword, KEY_STARDATE) == 0)
    {
       bStopParsing = DefineStardate(strParam);
    } else

    // the battle mode in a system ?
    if (strcmp(strKeyword, KEY_BATTLEMODE) == 0)
    {
       bStopParsing = DefineBattleMode(strParam);
    } else

    // ship change ?
    if (strcmp(strKeyword, KEY_SHIPCHANGE) == 0)
    {
       bStopParsing = DefineShipChange(strParam);
    } else

    // perfect cloak ?
    if (strcmp(strKeyword, KEY_PERFECTCLOAK) == 0)
    {
       bStopParsing = DefinePerfectCloak(strParam);
    } else

    // player has to stay until he cleared the system ?
    if (strcmp(strKeyword, KEY_WARPOUT) == 0)
    {
       bStopParsing = DefineWarpOut(strParam);
    } else

    // turn off collision checking for player ?
    if (strcmp(strKeyword, KEY_COLLISION) == 0)
    {
       bStopParsing = DefineCollision(strParam);
    } else

    // extended self destruct mode ?
    if (strcmp(strKeyword, KEY_SELFDESTRUCT) == 0)
    {
       bStopParsing = DefineSelfDestruct(strParam);
    } else

    // a title for the battle ?
    if (strcmp(strKeyword, KEY_TITLE) == 0)
    {
       bStopParsing = DefineTitle(strParam);
    } else

    // some title lines ?
    if (strcmp(strKeyword, KEY_TITLELINE) == 0)
    {
       bStopParsing = DefineTitleLine(strParam);
    } else

    // some epilog lines ?
    if (strcmp(strKeyword, KEY_EPILOGLINE) == 0)
    {
       bStopParsing = DefineEpilogLine(strParam);
    } else

    // some starting dialogs ?
    if (strcmp(strKeyword, KEY_STARTDIALOG) == 0)
    {
       bStopParsing = DefineStartDialog(strParam);
    } else

    // some ending dialogs ?
    if (strcmp(strKeyword, KEY_ENDDIALOG) == 0)
    {
       bStopParsing = DefineEndDialog(strParam);
    } else

    // some friend warp in dialogs ?
    if (strcmp(strKeyword, KEY_FRIENDWARPINDIALOG) == 0)
    {
       bStopParsing = DefineFriendWarpInDialog(strParam);
    } else

    // some enemy warp in dialogs ?
    if (strcmp(strKeyword, KEY_ENEMYWARPINDIALOG) == 0)
    {
       bStopParsing = DefineEnemyWarpInDialog(strParam);
    } else

    // some neutral warp in dialogs ?
    if (strcmp(strKeyword, KEY_NEUTRALWARPINDIALOG) == 0)
    {
       bStopParsing = DefineNeutralWarpInDialog(strParam);
    } else

    // some friend warp out dialogs ?
    if (strcmp(strKeyword, KEY_FRIENDWARPOUTDIALOG) == 0)
    {
       bStopParsing = DefineFriendWarpOutDialog(strParam);
    } else

    // some enemy warp out dialogs ?
    if (strcmp(strKeyword, KEY_ENEMYWARPOUTDIALOG) == 0)
    {
       bStopParsing = DefineEnemyWarpOutDialog(strParam);
    } else

    // some neutral warp out dialogs ?
    if (strcmp(strKeyword, KEY_NEUTRALWARPOUTDIALOG) == 0)
    {
       bStopParsing = DefineNeutralWarpOutDialog(strParam);
    } else

    // some friend join escort dialogs ?
    if (strcmp(strKeyword, KEY_JOINESCORTDIALOG) == 0)
    {
       bStopParsing = DefineFriendJoinEscortDialog(strParam);
    } else

    // some neutral to friend dialogs ?
    if (strcmp(strKeyword, KEY_NEUTRALTOFRIENDDIALOG) == 0)
    {
       bStopParsing = DefineNeutralToFriendDialog(strParam);
    } else

    // some neutral to enemy dialogs ?
    if (strcmp(strKeyword, KEY_NEUTRALTOENEMYDIALOG) == 0)
    {
       bStopParsing = DefineNeutralToEnemyDialog(strParam);
    } else

    // some jump gate appears dialogs ?
    if (strcmp(strKeyword, KEY_JUMPAPPEARSDIALOG) == 0)
    {
       bStopParsing = DefineJumpGateAppearsDialog(strParam);
    } else

    // some jump dialogs ?
    if (strcmp(strKeyword, KEY_JUMPDIALOG) == 0)
    {
       bStopParsing = DefineJumpDialog(strParam);
    } else

    // neutral behavior ?
    if (strcmp(strKeyword, KEY_NEUTRALBEHAVIOR) == 0)
    {
       bStopParsing = DefineNeutralBehavior(strParam);
    } else

    // warp disabled ?
    if (strcmp(strKeyword, KEY_NOWARP) == 0)
    {
       bStopParsing = DefineWarpDisable(strParam);
    } else

    // friend action ?
    if (strcmp(strKeyword, KEY_FRIENDACTION) == 0)
    {
       bStopParsing = DefineFriendAction(strParam);
    } else

    // enemy action ?
    if (strcmp(strKeyword, KEY_ENEMYACTION) == 0)
    {
       bStopParsing = DefineEnemyAction(strParam);
    } else

    // neutral action ?
    if (strcmp(strKeyword, KEY_NEUTRALACTION) == 0)
    {
       bStopParsing = DefineNeutralAction(strParam);
    } else

    // an empty line
    if ((strKeyword[0] == '\0') && (strParam[0] == '\0'))
    {
    } else

    // everything else is invalid
    {
      // unknown keyword ?
      if (strKeyword[0] != '\0')
      {
        ReportConfigError(ERR_INVALIDKEYWORD, TRUE);
      } else
      {
        ReportConfigError(ERR_INVALIDLINE, TRUE);
      }
      bStopParsing = TRUE;
    }


    // we had an error ?
    if (bStopParsing) 
    {
      ConfigFile.Close();
      return FALSE;
    }

    uParsedLine++;
  }

  ConfigFile.Close();

  if (pCurrentSystem != NULL)
  {
    // check some statistic battle mode related parameters for the last system
    if (!CheckStatBattleMode(pCurrentSystem, TRUE)) return FALSE;

    // check some other critical parameters for the last system
    if (!CheckOthers(pCurrentSystem)) return FALSE;
  }

  // and a final config check
  if (!FinalConfigCheck(FALSE)) return FALSE;
  
  return TRUE;
}

//-----------------------------------
// define a player
//-----------------------------------
BOOL CConfiguration::DefinePlayer(char *strParam)
{
  // do it
  return DefineShip(strParam, 0);
}

//-----------------------------------
// define an escort
//-----------------------------------
BOOL CConfiguration::DefineEscort(char *strParam)
{
  // do it
  return DefineShip(strParam, 1);
}

//-----------------------------------
// define a friend
//-----------------------------------
BOOL CConfiguration::DefineFriend(char *strParam)
{
  // do it
  return DefineShip(strParam, 2);
}

//-----------------------------------
// define an enemy
//-----------------------------------
BOOL CConfiguration::DefineEnemy(char *strParam)
{
  // do it
  return DefineShip(strParam, 3);
}

//-----------------------------------
// define a hulk
//-----------------------------------
BOOL CConfiguration::DefineHulk(char *strParam)
{
  // do it
  return DefineShip(strParam, 4);
}

//-----------------------------------
// define a neutral
//-----------------------------------
BOOL CConfiguration::DefineNeutral(char *strParam)
{
  // do it
  return DefineShip(strParam, 5);
}

//-----------------------------------
// define a jump gate
//-----------------------------------
BOOL CConfiguration::DefineJumpGate(char *strParam)
{
  // do it
  return DefineShip(strParam, 6);
}

//-----------------------------------
// define a ship
// (iWho: 0 = player, 1 = escort, 2 = friend, 3 = enemy, 4 = hulk, 5 = neutral, 6 = jump gate)
//-----------------------------------
BOOL CConfiguration::DefineShip(char *strParam, unsigned iWho)
{
  CShipData   *pShip;
  CShipList   *pShipList;
  BOOL        bShipNameSpecified;
  unsigned    i,j;
  char        strShipClass[MAX_QB2_LINELENGTH + 1];
  char        strShipName[MAX_QB2_LINELENGTH + 1];

  strShipName[0] = '\0';
  bShipNameSpecified = FALSE;

  // check if there is a second param, the ship name
  Tokenize(strShipClass, strParam, '=');
  // if so, isolate name and class
  if (strlen(strShipClass) != strlen(strParam))
  {
    // the shipname
    GetSubString(strShipName, strParam, strlen(strShipClass)+1, strlen(strParam)-1);
    StripString(strShipName, strShipName);
    // strip the shipclass
    StripString(strShipClass, strShipClass);
    // given name empty ?
    if (strShipName[0] == '\0')
    {
      ReportConfigError(ERR_EMPTYPARAM, TRUE);
      return TRUE;
    }
    bShipNameSpecified = TRUE;
  }
    
  // player already defined ?
  if ((iWho == 0) && (PlayerShip.IsSetShipClass()))
  {
    ReportConfigError(ERR_PLAYERREDEF, TRUE);
    return TRUE;
  }

  // system defined for enemy, friend, neutral and hulk
  if ((iWho > 1) && (pCurrentSystem == NULL))
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // jump gate already defined ?
  if ((iWho == 6) && (pCurrentSystem->GetJumpGate()->IsSetShipClass()))
  {
    ReportConfigError(ERR_JUMPGATEREDEF, TRUE);
    return TRUE;
  }

  // check the shipname and class
  if (!CheckParam(strShipClass)) return TRUE;
  if (bShipNameSpecified)
  {
    if (!CheckParam(strShipName)) return TRUE;
    // look if the shipname is unique
    // we only check the given names here, not the default names because the default
    // names may change if ships are inserted or deleted. So we do a complete name
    // check at the end of the configuration process
    if (!CheckShipName(strShipName, FALSE, TRUE, TRUE, &i, &j)) return TRUE;
  } 

  // put the ship in the proper list
  if (iWho == 0)
  {
    // player ship 
    PlayerShip.SetShipClass(strShipClass);
    if (bShipNameSpecified) PlayerShip.SetShipName(strShipName);
  } else
  if (iWho == 6)
  {
    // jump gate 
    pCurrentSystem->GetJumpGate()->SetShipClass(strShipClass);
    if (bShipNameSpecified) pCurrentSystem->GetJumpGate()->SetShipName(strShipName);
  } else
  {
    // create the new ship data object
    pShip = new CShipData;
    if (pShip == NULL)
    {
      ReportConfigError(ERR_OUTOFMEMORY, FALSE);
      return TRUE;
    }
    pShip->SetShipClass(strShipClass);
    if (bShipNameSpecified) pShip->SetShipName(strShipName);
    // add it to the proper list
    if (iWho == 1)
    {
      // the escort
      Escort.AddTail(pShip);
    } else
    if (iWho == 2)
    {
      // the friends
      pShipList = pCurrentSystem->GetFriends();
      pShipList->AddTail(pShip);
    } else    
    if (iWho == 3)
    {
      // the enemies
      pShipList = pCurrentSystem->GetEnemies();
      pShipList->AddTail(pShip);
    } else    
    if (iWho == 4)
    {
      // the hulks
      pShipList = pCurrentSystem->GetHulks();
      pShipList->AddTail(pShip);
    } else    
    {
      // the neutrals
      pShipList = pCurrentSystem->GetNeutrals();
      pShipList->AddTail(pShip);
    }
  }

  return FALSE;
}

//-----------------------------------
// define a system
//-----------------------------------
BOOL CConfiguration::DefineSystem(char *strParam)
{
  CSystemData *pSystem;

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // system exists already ?
  if (Systems.ExistsSystem(strParam))
  {
    ReportConfigError(ERR_SYSALREADYDEF, TRUE);
    return TRUE;
  }

  // too many systems ?
  if (Systems.GetCount() >= MAX_SYSTEMS)
  {
    ReportConfigError(ERR_TOOMANYSYS, TRUE);
    return TRUE;
  }

  // check some statistic battle mode related and other critical parameters for the previous system
  if (pCurrentSystem != NULL)
  {
    if (!CheckStatBattleMode(pCurrentSystem, TRUE)) return TRUE;
    if (!CheckOthers(pCurrentSystem)) return TRUE;
  }

  // add the new system
  pSystem = new CSystemData;
  if (pSystem == NULL)
  {
    ReportConfigError(ERR_OUTOFMEMORY, FALSE);
    return TRUE;
  }
  // set the name
  pSystem->SetSystemName(strParam);
  // add it to the list
  Systems.AddTail(pSystem);
  // set the current system pointer
  pCurrentSystem = pSystem;
  
  return FALSE;

}

//-----------------------------------
// define player position
//-----------------------------------
BOOL CConfiguration::DefinePlayerPosition(char *strParam)
{
  // do it
  return DefinePosition(strParam, 0);
}

//-----------------------------------
// define friend position
//-----------------------------------
BOOL CConfiguration::DefineFriendPosition(char *strParam)
{
  // do it
  return DefinePosition(strParam, 1);
}

//-----------------------------------
// define enemy position
//-----------------------------------
BOOL CConfiguration::DefineEnemyPosition(char *strParam)
{
  // do it
  return DefinePosition(strParam, 2);
}

//-----------------------------------
// define hulk position
//-----------------------------------
BOOL CConfiguration::DefineHulkPosition(char *strParam)
{
  // do it
  return DefinePosition(strParam, 3);
}

//-----------------------------------
// define neutral position
//-----------------------------------
BOOL CConfiguration::DefineNeutralPosition(char *strParam)
{
  // do it
  return DefinePosition(strParam, 4);
}

//-----------------------------------
// define jump gate position
//-----------------------------------
BOOL CConfiguration::DefineJumpGatePosition(char *strParam)
{
  // do it
  return DefinePosition(strParam, 5);
}

//-----------------------------------
// define a position
// (iWho: 0 = player, 1 = friend, 2 = enemy, 3 = hulk, 4 = neutral, 5 = jump gate)
//-----------------------------------
BOOL CConfiguration::DefinePosition(char *strParam, unsigned iWho)
{
  long x,y,z;
  char strPlanet[MAX_QB2_LINELENGTH + 1];

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // user already defined a position for this system
  if (iWho == 0)
  {
    // player
    if ((pCurrentSystem->IsSetPlayerPos()) || (pCurrentSystem->IsSetPlayerPosPlanet()))
    {
      ReportConfigError(ERR_PLAYERPOSREDEF, TRUE);
      return TRUE;
    }
  } else
  if (iWho == 1)
  {
    // friend
    if ((pCurrentSystem->IsSetFriendPos()) || (pCurrentSystem->IsSetFriendPosPlanet()))
    {
      ReportConfigError(ERR_FRIENDPOSREDEF, TRUE);
      return TRUE;
    }
  } else
  if (iWho == 2)
  {
    // enemy
    if ((pCurrentSystem->IsSetEnemyPos()) || (pCurrentSystem->IsSetEnemyPosPlanet()))
    {
      ReportConfigError(ERR_ENEMYPOSREDEF, TRUE);
      return TRUE;
    }
  } else
  if (iWho == 3)
  {
    // hulk
    if ((pCurrentSystem->IsSetHulkPos()) || (pCurrentSystem->IsSetHulkPosPlanet()))
    {
      ReportConfigError(ERR_HULKPOSREDEF, TRUE);
      return TRUE;
    }
  } else
  if (iWho == 4)
  {
    // neutral
    if ((pCurrentSystem->IsSetNeutralPos()) || (pCurrentSystem->IsSetNeutralPosPlanet()))
    {
      ReportConfigError(ERR_NEUTRALPOSREDEF, TRUE);
      return TRUE;
    }
  } else
  {
    // jump gate
    if ((pCurrentSystem->IsSetJumpGatePos()) || (pCurrentSystem->IsSetJumpGatePosPlanet()))
    {
      ReportConfigError(ERR_JUMPGATEPOSREDEF, TRUE);
      return TRUE;
    }
  }


  // position near a planet ?
  GetNthWordInString(strPlanet,strParam, 1);
  if (strcmp(strPlanet, SUBKEY_PLANET) == 0)
  {
    // yes, position near a planet
    // get the planet name
    GetSubString(strPlanet, strParam, strlen(SUBKEY_PLANET), strlen(strParam) - 1);
    // remove leading blanks
    StripString(strPlanet, strPlanet);
    // name ok ?
    if (!CheckParam(strPlanet)) return TRUE;

    // all is fine
    if (iWho == 0)
    {
      // player
      pCurrentSystem->SetPlayerPosPlanet(strPlanet);
    } else
    if (iWho == 1)
    {
      // friend
      pCurrentSystem->SetFriendPosPlanet(strPlanet);
    } else
    if (iWho == 2)
    {
      // enemy
      pCurrentSystem->SetEnemyPosPlanet(strPlanet);
    } else
    if (iWho == 3)
    {
      // hulk
      pCurrentSystem->SetHulkPosPlanet(strPlanet);
    } else
    if (iWho == 4)
    {
      // neutral
      pCurrentSystem->SetNeutralPosPlanet(strPlanet);
    } else
    {
      // jump gate
      pCurrentSystem->SetJumpGatePosPlanet(strPlanet);
    } 


  } else

  {
    // coordinates given
    // get x
    GetNthWordInString(strPlanet, strParam, 1);
    // check for errors
    if (!IsNumber(strPlanet))
    {
      ReportConfigError(ERR_INVALIDPOS, TRUE);
      return TRUE;
    }
    x = atol(strPlanet);
    // get y
    GetNthWordInString(strPlanet, strParam, 2);
    // check for errors
    if (!IsNumber(strPlanet))
    {
      ReportConfigError(ERR_INVALIDPOS, TRUE);
      return TRUE;
    }
    y = atol(strPlanet);
    // get z
    GetNthWordInString(strPlanet, strParam, 3);
    // check for errors
    if (!IsNumber(strPlanet))
    {
      ReportConfigError(ERR_INVALIDPOS, TRUE);
      return TRUE;
    }
    z = atol(strPlanet);

    // is this the last parameter ?
    GetNthWordInString(strPlanet, strParam, 4);
    if (strPlanet[0] != '\0')
    {
      ReportConfigError(ERR_TOOMANYPARAMS, TRUE);
      return TRUE;
    }

    // all is fine
    if (iWho == 0)
    {
      // player
      pCurrentSystem->SetPlayerPosX(x);
      pCurrentSystem->SetPlayerPosY(y);
      pCurrentSystem->SetPlayerPosZ(z);
    } else
    if (iWho == 1)
    {
      // friend
      pCurrentSystem->SetFriendPosX(x);
      pCurrentSystem->SetFriendPosY(y);
      pCurrentSystem->SetFriendPosZ(z);
    } else
    if (iWho == 2)
    {
      // enemy
      pCurrentSystem->SetEnemyPosX(x);
      pCurrentSystem->SetEnemyPosY(y);
      pCurrentSystem->SetEnemyPosZ(z);
    } else
    if (iWho == 3)
    {
      // hulk
      pCurrentSystem->SetHulkPosX(x);
      pCurrentSystem->SetHulkPosY(y);
      pCurrentSystem->SetHulkPosZ(z);
    } else
    if (iWho == 4)
    {
      // neutral
      pCurrentSystem->SetNeutralPosX(x);
      pCurrentSystem->SetNeutralPosY(y);
      pCurrentSystem->SetNeutralPosZ(z);
    } else
    {
      // jump gate
      pCurrentSystem->SetJumpGatePosX(x);
      pCurrentSystem->SetJumpGatePosY(y);
      pCurrentSystem->SetJumpGatePosZ(z);
    }


  }

  return FALSE;

}

//-----------------------------------
// define the stardate for a solar system
//-----------------------------------
BOOL CConfiguration::DefineStardate(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // user already defined a stardate for this system
  if (pCurrentSystem->IsSetStardate())
  {
    ReportConfigError(ERR_STARDATEREDEF, TRUE);
    return TRUE;
  }

  // check if it is a number
  if (!IsNumber(strParam))
  {
    ReportConfigError(ERR_INVSTARDATE, TRUE);
    return TRUE;
  }

  // set the stardate
  pCurrentSystem->SetStardate(atol(strParam));

  return FALSE;
}

//-----------------------------------
// define the jump time for a solar system
//-----------------------------------
BOOL CConfiguration::DefineJumpTime(char *strParam)
{
  long t;

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // user already defined a jump time for this system
  if (pCurrentSystem->IsSetJumpTime())
  {
    ReportConfigError(ERR_JUMPTIMEREDEF, TRUE);
    return TRUE;
  }

  // check if it is a number
  if (!IsNumber(strParam))
  {
    ReportConfigError(ERR_INVJUMPTIME, TRUE);
    return TRUE;
  }

  // set the time
  t = atol(strParam);
  if (t < 0) t = 0;
  pCurrentSystem->SetJumpTime(t);

  return FALSE;
}

//-----------------------------------
// define the jump destination for a solar system
//-----------------------------------
BOOL CConfiguration::DefineJumpDestination(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // user already defined a jump time for this system
  if (pCurrentSystem->IsSetJumpDestination())
  {
    ReportConfigError(ERR_JUMPDESTREDEF, TRUE);
    return TRUE;
  }

  // set the destination
  pCurrentSystem->SetJumpDestination(strParam);

  return FALSE;
}


//-----------------------------------
// define the jump options for a solar system
//-----------------------------------
BOOL CConfiguration::DefineJumpOptions(char *strParam)
{
  long iDist;
  char strParam1[MAX_QB2_LINELENGTH + 1];

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // hide the jump gate ?
  if (strcmp(strParam,SUBKEY_GATEHIDDEN) == 0)
  {
    // already defined ?
    if (pCurrentSystem->GetJumpOptions()->IsSetGateHidden())
    {
      ReportConfigError(ERR_GATEHIDDENREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->GetJumpOptions()->SetGateHidden();
  } else

  // allow only jumping to the destination system ?
  if (strcmp(strParam,SUBKEY_JUMPONLY) == 0)
  {
    // already defined ?
    if (pCurrentSystem->GetJumpOptions()->IsSetJumpOnly())
    {
      ReportConfigError(ERR_JUMPONLYREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->GetJumpOptions()->SetJumpOnly();
  } else

  // clear system after jump ?
  if (strcmp(strParam,SUBKEY_CLEARJUMP) == 0)
  {
    // already defined ?
    if (pCurrentSystem->GetJumpOptions()->IsSetClearJump())
    {
      ReportConfigError(ERR_CLEARJUMPREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->GetJumpOptions()->SetClearJump();
  } else

  // no spin after jump ?
  if (strcmp(strParam,SUBKEY_NOSPIN) == 0)
  {
    // already defined ?
    if (pCurrentSystem->GetJumpOptions()->IsSetNoSpin())
    {
      ReportConfigError(ERR_NOSPINREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->GetJumpOptions()->SetNoSpin();
  } else

  // show activation ?
  if (strcmp(strParam,SUBKEY_SHOWACT) == 0)
  {
    // already defined ?
    if (pCurrentSystem->GetJumpOptions()->IsSetShowActivation())
    {
      ReportConfigError(ERR_SHOWJUMPREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->GetJumpOptions()->SetShowActivation();
  } else

  // show activation ?
  if (strcmp(strParam,SUBKEY_ACTONHAIL) == 0)
  {
    // already defined ?
    if (pCurrentSystem->GetJumpOptions()->IsSetActOnHail())
    {
      ReportConfigError(ERR_HAILJUMPREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->GetJumpOptions()->SetActOnHail();
  } else
  {

    // modify activation distance
    GetNthWordInString(strParam1,strParam, 1);
    if (strcmp(strParam1,SUBKEY_ACTDISTANCE) == 0)
    {
      // already defined ?
      if (pCurrentSystem->GetJumpOptions()->IsSetActivationDistance())
      {
        ReportConfigError(ERR_DISTJUMPREDEF, TRUE);
        return TRUE;
      }

      // get the new distance
      GetNthWordInString(strParam1,strParam, 2);

      // check if it is a number
      if (!IsNumber(strParam1))
      {
        ReportConfigError(ERR_INVPARAM, TRUE);
        return TRUE;
      }
      iDist = atol(strParam1);
      if (iDist < 0) iDist = 0;

      // too many parameters ?
      GetNthWordInString(strParam1, strParam, 3);
      if (strParam1[0] != '\0')
      {
        ReportConfigError(ERR_TOOMANYPARAMS, TRUE);
        return TRUE;
      }

      // set the new distance
      pCurrentSystem->GetJumpOptions()->SetActivationDistance(iDist);
    } else

    // everything else is error
    {
      ReportConfigError(ERR_INVPARAM, TRUE);
      return TRUE;
    }
  }  

  return FALSE;
}

//-----------------------------------
// define the battle mode in a solar system
//-----------------------------------
BOOL CConfiguration::DefineBattleMode(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // user already defined a battle mode for this system
  if (pCurrentSystem->IsSetBattleMode())
  {
    ReportConfigError(ERR_BATTLEMODEREDEF, TRUE);
    return TRUE;
  }

  // set the battle mode
  if (strcmp(strParam,SUBKEY_STATBATTLE) == 0)
  {
    pCurrentSystem->SetBattleMode(StatisticBattle);
  } else
  if (strcmp(strParam,SUBKEY_AIBATTLE) == 0)
  {
    pCurrentSystem->SetBattleMode(AIBattle);
  } else
  {
    ReportConfigError(ERR_INVBATTLEMODE, TRUE);
    return TRUE;
  }

  return FALSE;
}

//-----------------------------------
// define the posibility of a ship change
//-----------------------------------
BOOL CConfiguration::DefineShipChange(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // user already defined this
  if (GeneralMode.IsSetShipChangeMode())
  {
    ReportConfigError(ERR_SHIPCHANGEREDEF, TRUE);
    return TRUE;
  }

  // ship change off ?
  if (strcmp(strParam,SUBKEY_OFF) == 0)
  {
    GeneralMode.SetShipChangeMode(NoShipChange);
  } else

  // ship change on ?
  if (strcmp(strParam,SUBKEY_ON) == 0)
  {
    GeneralMode.SetShipChangeMode(ManualShipChange);
  } else

  // ship change auto
  if (strcmp(strParam,SUBKEY_AUTO) == 0)
  {
    GeneralMode.SetShipChangeMode(AutoShipChange);
  } else

  // everything else is error
  {
    ReportConfigError(ERR_INVPARAM, TRUE);
    return TRUE;
  }

  return FALSE;
}

//-----------------------------------
// define the posibility of a perfect cloak
//-----------------------------------
BOOL CConfiguration::DefinePerfectCloak(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // user already defined this
  if (GeneralMode.IsSetPerfectCloakMode())
  {
    ReportConfigError(ERR_PCLOAKREDEF, TRUE);
    return TRUE;
  }

  // perfect cloak off ?
  if (strcmp(strParam,SUBKEY_OFF) == 0)
  {
    GeneralMode.SetPerfectCloakMode(PerfectCloakOff);
  } else

  // perfect cloak on ?
  if (strcmp(strParam,SUBKEY_ON) == 0)
  {
    GeneralMode.SetPerfectCloakMode(PerfectCloakOn);
  } else

  // everything else is error
  {
    ReportConfigError(ERR_INVPARAM, TRUE);
    return TRUE;
  }

  return FALSE;
}

//-----------------------------------
// define when the player is allowed to leave the system
//-----------------------------------
BOOL CConfiguration::DefineWarpOut(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // user already defined this
  if (GeneralMode.IsSetWarpOutMode())
  {
    ReportConfigError(ERR_MUSTSTAYREDEF, TRUE);
    return TRUE;
  }

  // sequential system order ?
  if (strcmp(strParam,SUBKEY_SEQUENTIAL) == 0)
  {
    GeneralMode.SetWarpOutMode(Sequential);
  } else

  // must stay until cleared?
  if (strcmp(strParam,SUBKEY_CLEARED) == 0)
  {
    GeneralMode.SetWarpOutMode(WhenCleared);
  } else

  // may leave anytime
  if (strcmp(strParam,SUBKEY_ANYTIME) == 0)
  {
    GeneralMode.SetWarpOutMode(Anytime);
  } else

  // everything else is error
  {
    ReportConfigError(ERR_INVPARAM, TRUE);
    return TRUE;
  }

  return FALSE;
}

//-----------------------------------
// define collision checking for the playership and the rest
//-----------------------------------
BOOL CConfiguration::DefineCollision(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // user already defined this
  if (GeneralMode.IsSetCollisionMode())
  {
    ReportConfigError(ERR_COLLISIONREDEF, TRUE);
    return TRUE;
  }

  // collision on ?
  if (strcmp(strParam,SUBKEY_ON) == 0)
  {
    GeneralMode.SetCollisionMode(CollisionsOn);
  } else

  // collision off ?
  if (strcmp(strParam,SUBKEY_OFF) == 0)
  {
    GeneralMode.SetCollisionMode(CollisionsOff);
  } else

  // collision off for all ?
  if (strcmp(strParam,SUBKEY_OFFALL) == 0)
  {
    GeneralMode.SetCollisionMode(CollisionsOffForAll);
  } else

  // everything else is error
  {
    ReportConfigError(ERR_INVPARAM, TRUE);
    return TRUE;
  }

  return FALSE;
}


//-----------------------------------
// define the self destruct options for the playership
//-----------------------------------
BOOL CConfiguration::DefineSelfDestruct(char *strParam)
{
  char    strOption[MAX_QB2_LINELENGTH + 1];
  char    strOption1[MAX_QB2_LINELENGTH + 1];
  char    strCode[MAX_QB2_LINELENGTH + 1];
  long    dist;

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // user already defined this
  if (SelfDestruct.IsSetMode())
  {
    ReportConfigError(ERR_SELFDESTRUCTREDEF, TRUE);
    return TRUE;
  }

  // check if there is a second param, the destruct code
  Tokenize(strOption, strParam, '=');
  strCode[0] = '\0';
  // if so, isolate option and code
  if (strlen(strOption) != strlen(strParam))
  {
    // the code
    GetSubString(strCode, strParam, strlen(strOption)+1, strlen(strParam)-1);
    StripString(strCode, strCode);
    // strip the options
    StripString(strOption, strOption);
    // given code not empty ?
    if (strCode[0] == '\0')
    {
      ReportConfigError(ERR_EMPTYPARAM, TRUE);
      return TRUE;
    }
  }

  // store the code
  SelfDestruct.SetCode(strCode);

  // evaluate the options

  // standard self destruct, no overkill ?
  if (strcmp(strOption,SUBKEY_STANDARD) == 0)
  {
    SelfDestruct.SetMode(StandardDestruct);
    SelfDestruct.SetOverkillDistance(0);
  } else
  {
    // overkill or selective overkill destruct, get option and distance
    GetNthWordInString(strOption1, strOption, 1);

    // overkill ?
    if (strcmp(strOption1,SUBKEY_OVERKILL) == 0)
    {
      SelfDestruct.SetMode(OverkillDestruct);
    } else

    // selective overkill ?
    if (strcmp(strOption1,SUBKEY_SELOVERKILL) == 0)
    {
      SelfDestruct.SetMode(SelectiveOverkillDestruct);
    } else

    // everything else is an error
    {
      ReportConfigError(ERR_INVSELFDESTRUCT, TRUE);
      return TRUE;
    }
  }

  // set the overkill distance
  if ((SelfDestruct.GetMode() == OverkillDestruct) ||
      (SelfDestruct.GetMode() == SelectiveOverkillDestruct))
  {
    GetNthWordInString(strOption1, strOption, 2);
    if (!IsNumber(strOption1))
    {
      ReportConfigError(ERR_INVSELFDESTRUCT, TRUE);
      return TRUE;
    }
    dist = atol(strOption1);
    if (dist <0) dist = 0;
    SelfDestruct.SetOverkillDistance(dist);
  }

  // too many parameters
  GetNthWordInString(strOption1, strOption, 3);
  if (strOption1[0] != '\0')
  {
    ReportConfigError(ERR_TOOMANYPARAMS, TRUE);
    return TRUE;
  }

  return FALSE;

}

//-----------------------------------
// define a battle title
//-----------------------------------
BOOL CConfiguration::DefineTitle(char *strParam)
{
  char  strParam1[MAX_QB2_LINELENGTH + 1];

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // user already defined this
  if (Title.IsSetTitle())
  {
    ReportConfigError(ERR_TITLEREDEF, TRUE);
    return TRUE;
  }

  // check line length
  if (strlen(strParam) > MAX_TITLE)
  {
    ReportConfigError(ERR_LONGTITLE, TRUE);
    return TRUE;
  }

  // check # (list delimiter not allowed)
  Tokenize(strParam1, strParam, '#');
  if (strlen(strParam1) != strlen(strParam))
  {
    ReportConfigError(ERR_INVALIDFIS, TRUE);
    return TRUE;
  }

  // set the title
  Title.SetTitle(strParam);

  return FALSE;
}

//-----------------------------------
// define some title lines for a battle title
//-----------------------------------
BOOL CConfiguration::DefineTitleLine(char *strParam)
{
  // do it
  return DefineTitleEpilogLine(strParam, 0);
}

//-----------------------------------
// define some epilog lines for a battle epilog
//-----------------------------------
BOOL CConfiguration::DefineEpilogLine(char *strParam)
{
  // do it
  return DefineTitleEpilogLine(strParam, 1);
}

//-----------------------------------
// define some title or epilog lines 
// (iWhat = 0: title line, iWhat = 1: eplog line)
//-----------------------------------
BOOL CConfiguration::DefineTitleEpilogLine(char *strParam, unsigned iWhat)
{
  char     strParam1[MAX_QB2_LINELENGTH + 1];
  char     strVoice[MAX_QB2_LINELENGTH + 1];
  CStrList *TextLines;
  CStrList *TextVoices;
  CString  *TextLine;
  CString  *TextVoice;

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // isolate the optional voice file
  strVoice[0] = '\0';
  if (strParam[0] == '[')
  {
    // look for closing bracket
    Tokenize(strVoice, strParam, ']');
    // is there a voice file ?
    if (strlen(strVoice) == strlen(strParam))
    {
      // no
      strVoice[0] = '\0';
    } else
    {
      // yes we have a voice file, remove it from the input string
      GetSubString(strParam, strParam, strlen(strVoice)+1, strlen(strParam)-1);
      StripString(strParam, strParam);
      // remove the opening bracket from the voice file
      GetSubString(strVoice, strVoice, 1, strlen(strVoice)-1);
      StripString(strVoice, strVoice);
    }
  }

  // check line length
  if (strlen(strParam) > MAX_TITLELINE)
  {
    ReportConfigError(ERR_LONGTITLELINE, TRUE);
    return TRUE;
  }

  // check empty line
  if (strParam[0] == '\0')
  {
    ReportConfigError(ERR_EMPTYTITLELINE, TRUE);
    return TRUE;
  }

  // check # (list delimiter not allowed)
  Tokenize(strParam1, strParam, '#');
  if (strlen(strParam1) != strlen(strParam))
  {
    ReportConfigError(ERR_INVALIDFIS, TRUE);
    return TRUE;
  }
  Tokenize(strParam1, strVoice, '#');
  if (strlen(strParam1) != strlen(strVoice))
  {
    ReportConfigError(ERR_INVALIDFIS, TRUE);
    return TRUE;
  }

  // create the title or epilog line and voice
  TextLine = new CString;
  TextVoice = new CString;
  if ((TextLine == NULL) || (TextVoice == NULL))
  {
    ReportConfigError(ERR_OUTOFMEMORY, FALSE);
    return TRUE;
  }
  *TextLine = strParam;
  *TextVoice = strVoice;

  // add it to the list
  if (iWhat == 0)
  {
    // the title lines list
    TextLines = Title.GetTitleLines();
    TextVoices = Title.GetTitleVoices();
  } else
  {
    // the epilog lines list
    TextLines = Epilog.GetEpilogLines();
    TextVoices = Epilog.GetEpilogVoices();
  }
  // add it
  TextLines->AddTail((CObject *)TextLine);
  TextVoices->AddTail((CObject *)TextVoice);

  return FALSE;
}

//-----------------------------------
// define a starting dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineStartDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,0);
}

//-----------------------------------
// define an ending dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineEndDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,1);
}

//-----------------------------------
// define a friend warp in dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineFriendWarpInDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,2);
}

//-----------------------------------
// define an enemy warp in dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineEnemyWarpInDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,3);
}

//-----------------------------------
// define an neutral warp in dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineNeutralWarpInDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,4);
}

//-----------------------------------
// define a friend warp out dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineFriendWarpOutDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,5);
}

//-----------------------------------
// define an enemy warp out dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineEnemyWarpOutDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,6);
}

//-----------------------------------
// define an neutral warp out dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineNeutralWarpOutDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,7);
}

//-----------------------------------
// define a friend join escort dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineFriendJoinEscortDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,8);
}

//-----------------------------------
// define a neutral to friend dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineNeutralToFriendDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,9);
}

//-----------------------------------
// define a neutral to enemy dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineNeutralToEnemyDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,10);
}

//-----------------------------------
// define a jump gate appears dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineJumpGateAppearsDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,11);
}

//-----------------------------------
// define a jump dialog for a system
//-----------------------------------
BOOL CConfiguration::DefineJumpDialog(char *strParam)
{

  // do it
  return DefineDialog(strParam,12);
}

//-----------------------------------
// define a dialog for a system
// (iWhat = 0: starting dialog, 1: ending dialog, 2: friend warp in dialog,
//          3: enemy warp in dialog, 4: neutral warp in dialog,
//          5: friend warp out dialog, 6: enemy warp out dialog,
//          7: neutral warp out dialog, 8: friend join escort dialog,
//          9: neutral to friend dialog, 10: neutral to enemy dialog,
//          11: jump gate appears dialog, 12: jump dialog)
//-----------------------------------
BOOL CConfiguration::DefineDialog(char *strParam, unsigned iWhat)
{
  char        strParam1[MAX_QB2_LINELENGTH + 1];
  char        strSpeaker[MAX_QB2_LINELENGTH + 1];
  char        strLine[MAX_QB2_LINELENGTH + 1];
  char        strVoice[MAX_QB2_LINELENGTH + 1];
  CDialogData *pDialog;
  CDialogList *pDialogList;

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // isolate the optional voice file
  strVoice[0] = '\0';
  if (strParam[0] == '[')
  {
    // look for closing bracket
    Tokenize(strVoice, strParam, ']');
    // is there a voice file ?
    if (strlen(strVoice) == strlen(strParam))
    {
      // no
      strVoice[0] = '\0';
    } else
    {
      // yes we have a voice file, remove it from the input string
      GetSubString(strParam, strParam, strlen(strVoice)+1, strlen(strParam)-1);
      StripString(strParam, strParam);
      // remove the opening bracket from the voice file
      GetSubString(strVoice, strVoice, 1, strlen(strVoice)-1);
      StripString(strVoice, strVoice);
    }
  }

  // isolate the speaker
  Tokenize(strSpeaker, strParam, ':');

  // is there a speaker ?
  if (strlen(strSpeaker) == strlen(strParam))
  { 
    ReportConfigError(ERR_NOSPEAKER, TRUE);
    return TRUE;
  }

  // get the dialog line
  GetSubString(strLine, strParam, strlen(strSpeaker)+1, strlen(strParam)-1);
  StripString(strLine, strLine);
  // strip the speaker
  StripString(strSpeaker, strSpeaker);

  // given names empty or too long ?
  if (strSpeaker[0] == '\0')
  { 
    ReportConfigError(ERR_NOSPEAKER, TRUE);
    return TRUE;
  }
  if (strLine[0] == '\0')
  { 
    ReportConfigError(ERR_NODIALOG, TRUE);
    return TRUE;
  }
  if (strlen(strSpeaker) > MAX_SPEAKER)
  { 
    ReportConfigError(ERR_LONGSPEAKER, TRUE);
    return TRUE;
  }
  if (strlen(strLine) > MAX_DIALINE)
  { 
    ReportConfigError(ERR_LONGDIALOG, TRUE);
    return TRUE;
  }

  // check # (list delimiter not allowed)
  Tokenize(strParam1, strSpeaker, '#');
  if (strlen(strParam1) != strlen(strSpeaker))
  {
    ReportConfigError(ERR_INVALIDFIS, TRUE);
    return TRUE;
  }
  Tokenize(strParam1, strLine, '#');
  if (strlen(strParam1) != strlen(strLine))
  {
    ReportConfigError(ERR_INVALIDFIS, TRUE);
    return TRUE;
  }
  Tokenize(strParam1, strVoice, '#');
  if (strlen(strParam1) != strlen(strVoice))
  {
    ReportConfigError(ERR_INVALIDFIS, TRUE);
    return TRUE;
  }

  // all is fine, create the object
  pDialog = new CDialogData;
  if (pDialog == NULL)
  {
    ReportConfigError(ERR_OUTOFMEMORY, FALSE);
    return TRUE;
  }
  pDialog->SetSpeaker(strSpeaker);
  pDialog->SetDialogLine(strLine);
  pDialog->SetVoiceFile(strVoice);

  // add it to the list for the current system
  if (iWhat == 0)
  {
    pDialogList = pCurrentSystem->GetStartingDialogs();
  } else
  if (iWhat == 1)
  {
    pDialogList = pCurrentSystem->GetEndingDialogs();
  } else
  if (iWhat == 2)
  {
    pDialogList = pCurrentSystem->GetFriendWarpInDialogs();
  } else
  if (iWhat == 3)
  {
    pDialogList = pCurrentSystem->GetEnemyWarpInDialogs();
  } else
  if (iWhat == 4)
  {
    pDialogList = pCurrentSystem->GetNeutralWarpInDialogs();
  } else
  if (iWhat == 5)
  {
    pDialogList = pCurrentSystem->GetFriendWarpOutDialogs();
  } else
  if (iWhat == 6)
  {
    pDialogList = pCurrentSystem->GetEnemyWarpOutDialogs();
  } else
  if (iWhat == 7)
  {
    pDialogList = pCurrentSystem->GetNeutralWarpOutDialogs();
  } else
  if (iWhat == 8)
  {
    pDialogList = pCurrentSystem->GetFriendJoinEscortDialogs();
  } else
  if (iWhat == 9)
  {
    pDialogList = pCurrentSystem->GetNeutralToFriendDialogs();
  } else
  if (iWhat == 10)
  {
    pDialogList = pCurrentSystem->GetNeutralToEnemyDialogs();
  } else
  if (iWhat == 11)
  {
    pDialogList = pCurrentSystem->GetJumpGateAppearsDialogs();
  } else
  if (iWhat == 12)
  {
    pDialogList = pCurrentSystem->GetJumpDialogs();
  }

  // do it
  pDialogList->AddTail(pDialog);

  return FALSE;
}

//-----------------------------------
// define the behavior of the neutrals in a system
//-----------------------------------
BOOL CConfiguration::DefineNeutralBehavior(char *strParam)
{
  char  strParam1[MAX_QB2_LINELENGTH + 1];
  unsigned long t;

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // user already defined a neutral for this system
  if (pCurrentSystem->IsSetNeutralBehavior())
  {
    ReportConfigError(ERR_NEUTBEHAVREDEF, TRUE);
    return TRUE;
  }

  // set the behavior
  if (strcmp(strParam,SUBKEY_ENEMYWHENATTACK) == 0)
  {
    pCurrentSystem->SetNeutralBehavior(EnemyWhenAttacked);
  } else
  {
    GetNthWordInString(strParam1, strParam, 1);
    if (strcmp(strParam1,SUBKEY_ENEMYAFTER) == 0)
    {
      pCurrentSystem->SetNeutralBehavior(EnemyAfter);
    } else
    if (strcmp(strParam1,SUBKEY_FRIENDAFTER) == 0)
    {
      pCurrentSystem->SetNeutralBehavior(FriendAfter);
    } else
    if (strcmp(strParam1,SUBKEY_RANDOM) == 0)
    {
      pCurrentSystem->SetNeutralBehavior(RandomBehavior);
    } else
    {
      ReportConfigError(ERR_INVNEUTBEHAV, TRUE);
      return TRUE;
    }
  }

  // set the time
  if ((pCurrentSystem->GetNeutralBehavior() == EnemyAfter) ||
      (pCurrentSystem->GetNeutralBehavior() == FriendAfter))
  {
    GetNthWordInString(strParam1, strParam, 2);
    if (!IsNumber(strParam1))
    {
      ReportConfigError(ERR_INVNEUTBEHAV, TRUE);
      return TRUE;
    }
    t = atol(strParam1);
    pCurrentSystem->SetNeutralBehaviorTime(t);
  }

  // too many parameters
  GetNthWordInString(strParam1, strParam, 3);
  if (strParam1[0] != '\0')
  {
    ReportConfigError(ERR_TOOMANYPARAMS, TRUE);
    return TRUE;
  }

  return FALSE;
}


//-----------------------------------
// disable warp for some groups
//-----------------------------------
BOOL CConfiguration::DefineWarpDisable(char *strParam)
{

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // disable enemy warp ?
  if (strcmp(strParam,SUBKEY_ENEMY) == 0)
  {
    // already defined ?
    if (pCurrentSystem->IsSetEnemyWarpDisable())
    {
      ReportConfigError(ERR_ENEMYWARPREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->SetEnemyWarpDisable(TRUE);
  } else

  // disable friend warp ?
  if (strcmp(strParam,SUBKEY_FRIEND) == 0)
  {
    // already defined ?
    if (pCurrentSystem->IsSetFriendWarpDisable())
    {
      ReportConfigError(ERR_FRIENDWARPREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->SetFriendWarpDisable(TRUE);
  } else

  // disable neutral warp ?
  if (strcmp(strParam,SUBKEY_NEUTRAL) == 0)
  {
    // already defined ?
    if (pCurrentSystem->IsSetNeutralWarpDisable())
    {
      ReportConfigError(ERR_NEUTRALWARPREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->SetNeutralWarpDisable(TRUE);
  } else

  // disable player warp ?
  if (strcmp(strParam,SUBKEY_PLAYER) == 0)
  {
    // already defined ?
    if (pCurrentSystem->IsSetPlayerWarpDisable())
    {
      ReportConfigError(ERR_PLAYERWARPREDEF, TRUE);
      return TRUE;
    }
    pCurrentSystem->SetPlayerWarpDisable(TRUE);
  } else

  // everything else is error
  {
    ReportConfigError(ERR_INVPARAM, TRUE);
    return TRUE;
  }

  return FALSE;
}


//-----------------------------------
// define a friend action
//-----------------------------------
BOOL CConfiguration::DefineFriendAction(char *strParam)
{

  // do it
  return DefineAction(strParam, 0);
}

//-----------------------------------
// define an enemy action
//-----------------------------------
BOOL CConfiguration::DefineEnemyAction(char *strParam)
{

  // do it
  return DefineAction(strParam, 1);
}

//-----------------------------------
// define a neutral action
//-----------------------------------
BOOL CConfiguration::DefineNeutralAction(char *strParam)
{

  // do it
  return DefineAction(strParam, 2);
}

//-----------------------------------
// define an action for a groups
// (iWho = 0: friends  1: enemies  2: neutrals)
//-----------------------------------
BOOL CConfiguration::DefineAction(char *strParam, unsigned iWho)
{
  long iTime = 0;
  char strParam1[MAX_QB2_LINELENGTH + 1];

  // check the param
  if (!CheckParam(strParam)) return TRUE;

  // valid system
  if (pCurrentSystem == NULL)
  {
    ReportConfigError(ERR_SYSEXPECTED, TRUE);
    return TRUE;
  }

  // warpin action
  GetNthWordInString(strParam1,strParam, 1);
  if (strcmp(strParam1, SUBKEY_WARPIN) == 0)
  {

    // get the time
    GetNthWordInString(strParam1, strParam, 2);
    // check the time
    if (!IsNumber(strParam1))
    {
      ReportConfigError(ERR_INVACTION, TRUE);
      return TRUE;
    }
    iTime = atol(strParam1);

    // who ?
    if (iWho == 0)
    {
      // friends
      // action already defined ?
      if (pCurrentSystem->IsSetFriendWarpInTime())
      {
        ReportConfigError(ERR_FRIENDWARPINREDEF, TRUE);
        return TRUE;
      }
      // set the action
      pCurrentSystem->SetFriendWarpInTime(iTime);
    } else

    if (iWho == 1)
    {
      // enemies
      // action already defined ?
      if (pCurrentSystem->IsSetEnemyWarpInTime())
      {
        ReportConfigError(ERR_ENEMYWARPINREDEF, TRUE);
        return TRUE;
      }
      // set the action
      pCurrentSystem->SetEnemyWarpInTime(iTime);
    } else

    {
      // neutrals
      // action already defined ?
      if (pCurrentSystem->IsSetNeutralWarpInTime())
      {
        ReportConfigError(ERR_NEUTRALWARPINREDEF, TRUE);
        return TRUE;
      }
      // set the action
      pCurrentSystem->SetNeutralWarpInTime(iTime);
    }
  } else

  // warpout action
  if (strcmp(strParam1, SUBKEY_WARPOUT) == 0)
  {

    // get the time
    GetNthWordInString(strParam1, strParam, 2);
    // check the time
    if (!IsNumber(strParam1))
    {
      // check for sub-keyword 'cleared' instead of a time
      if (strcmp(strParam1, SUBKEY_TCLEARED) == 0)
      {
        // warp-out when cleared is indicated by time = -1
        iTime = -1;
      } else
      {
        // not ok
        ReportConfigError(ERR_INVACTION, TRUE);
        return TRUE;
      }
    } else
    {
      iTime = atol(strParam1);
    }

    // who ?
    if (iWho == 0)
    {
      // friends
      // action already defined ?
      if (pCurrentSystem->IsSetFriendWarpOutTime())
      {
        ReportConfigError(ERR_FRIENDWARPOUTREDEF, TRUE);
        return TRUE;
      }
      // set the action
      pCurrentSystem->SetFriendWarpOutTime(iTime);
    } else

    if (iWho == 1)
    {
      // enemies
      // action already defined ?
      if (pCurrentSystem->IsSetEnemyWarpOutTime())
      {
        ReportConfigError(ERR_ENEMYWARPOUTREDEF, TRUE);
        return TRUE;
      }
      // set the action
      pCurrentSystem->SetEnemyWarpOutTime(iTime);
    } else

    {
      // neutrals
      // action already defined ?
      if (pCurrentSystem->IsSetNeutralWarpOutTime())
      {
        ReportConfigError(ERR_NEUTRALWARPOUTREDEF, TRUE);
        return TRUE;
      }
      // set the action
      pCurrentSystem->SetNeutralWarpOutTime(iTime);
    }
  } else

  // join escort action
  if (strcmp(strParam1, SUBKEY_JOINESCORT) == 0)
  {

    // get the time
    GetNthWordInString(strParam1, strParam, 2);
    // check the time
    if (!IsNumber(strParam1))
    {
      // check for sub-keyword 'cleared' instead of a time
      if (strcmp(strParam1, SUBKEY_TCLEARED) == 0)
      {
        // warp-out when cleared is indicated by time = -1
        iTime = -1;
      } else
      {
        // not ok
        ReportConfigError(ERR_INVACTION, TRUE);
        return TRUE;
      }
    } else
    {
      iTime = atol(strParam1);
    }

    // this action is only available for friends ?
    if (iWho != 0)
    {
      ReportConfigError(ERR_ONLYFRIENDSJOIN, TRUE);
      return TRUE;
    }

    // action already defined ?
    if (pCurrentSystem->IsSetFriendJoinEscortTime())
    {
      ReportConfigError(ERR_JOINESCORTREDEF, TRUE);
      return TRUE;
    }
    // set the action
    pCurrentSystem->SetFriendJoinEscortTime(iTime);
  } else

  // invalid action
  {
    ReportConfigError(ERR_INVACTION, TRUE);
    return TRUE;
  }

  // check for more errors
  // not too many params ?
  GetNthWordInString(strParam1, strParam, 3);
  if (strParam1[0] != '\0')
  {
    ReportConfigError(ERR_TOOMANYPARAMS, TRUE);
    return TRUE;
  }

  return FALSE;

}

//-----------------------------------
// writes a line to the config file, limits line length
//-----------------------------------
BOOL CConfiguration::WriteCfgLine(const char *strLine)
{
  char strLimitedLine[MAX_QB2_LINELENGTH + 1];

  // limit the line length
  GetSubString(strLimitedLine, strLine, 0, MAX_QB2_LINELENGTH - 1);
  // write it
  if (!ConfigFile.WriteLine(strLimitedLine))
  {
    ReportConfigError(ERR_CFGWRITE, FALSE);
    return FALSE;
  }
  return TRUE;
}

//-----------------------------------
// sets up an output line containing a ship class and name
//-----------------------------------
void CConfiguration::OutputShipClassAndName(CString *line, CShipData *ship)
{
  // first set the class
  *line = ship->GetShipClass();
  // do we have a name for the ship
  if (ship->IsSetShipName())
  {
    // yes, add it
    *line = *line + " = ";
    *line = *line + ship->GetShipName();
  }

}


//-----------------------------------
// write a configuration
//-----------------------------------
BOOL CConfiguration::Write(char *ConfigName)
{
  CString     strConfigFileName;
  CSystemData *pSystem;
  POSITION    pos;

  // get the name of the config file
  strConfigFileName = strPathToKA + QB2_PATH + ConfigName + ".qb2";
  strConfigName = ConfigName;

  // create the new config file
  if (!ConfigFile.Create(strConfigFileName))
  {
    // an unexpected error
    ReportConfigError(ERR_CFGCREATE, FALSE);
    return FALSE;
  }

  // write the header
  if (!WriteHeader())
  {
    ConfigFile.Close();
    return FALSE;
  }

  // write the title and the titlelines
  if (!WriteTitleAndTitleLines())
  {
    ConfigFile.Close();
    return FALSE;
  }

  // write the epiloglines
  if (!WriteEpilogLines())
  {
    ConfigFile.Close();
    return FALSE;
  }

  // write the player and his escort
  if (!WritePlayerAndEscort())
  {
    ConfigFile.Close();
    return FALSE;
  }

  // write the general parameters
  if (!WriteGeneral())
  {
    ConfigFile.Close();
    return FALSE;
  }

  // write the self destruct mode
  if (!WriteSelfDestruct())
  {
    ConfigFile.Close();
    return FALSE;
  }

  // loop through all systems
  for (pos = Systems.GetHeadPosition(); pos != NULL; )
  {
    // iterate through systems
    pSystem = (CSystemData *)Systems.GetNext(pos);

    // write the system name
    if (!WriteSystemName(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the stardate
    if (!WriteStardate(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the friends in the system
    if (!WriteFriends(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }
  
    // write the enemies in the system
    if (!WriteEnemies(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }
  
    // write the neutrals in the system
    if (!WriteNeutrals(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }
  
    // write the hulks in the system
    if (!WriteHulks(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }
  
    // write the jump gate in the system
    if (!WriteJumpGate(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }
  
    // write the jump time in the system
    if (!WriteJumpTime(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the jump destination in the system
    if (!WriteJumpDestination(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the jump options in the system
    if (!WriteJumpOptions(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the positions in the system
    if (!WritePositions(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the actions in the system
    if (!WriteActions(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the neutral behavior in the system
    if (!WriteNeutralBehavior(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the dialogs in the system
    if (!WriteDialogs(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the warp disables in the system
    if (!WriteWarpDisable(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

    // write the battle mode in the system
    if (!WriteBattleMode(pSystem))
    {
      ConfigFile.Close();
      return FALSE;
    }

  }

  ConfigFile.Close();
 
  return TRUE;
}

//-----------------------------------
// write the header of the config file
//-----------------------------------
BOOL CConfiguration::WriteHeader()
{
  CString  *pCStr, CStr;
  POSITION pos;

  // write the magic and version number
  if (!WriteCfgLine(QB2_MAGIC)) return FALSE;
  if (!WriteCfgLine(QB2_VERSION4)) return FALSE;

  // prepare remark characters
  CStr = REMARK;

  // write the initial remarks (including qb2edit brand)
  for (pos = Remarks.GetHeadPosition(); pos != NULL; )
  {
    // iterate through remarks
    pCStr = (CString *)Remarks.GetNext(pos);
    // add remark characters and write it
    if (!WriteCfgLine(CStr + *pCStr)) return FALSE;
  }

  return TRUE;
}

//-----------------------------------
// write the title and the titlelines
//-----------------------------------
BOOL CConfiguration::WriteTitleAndTitleLines()
{
  CString  *pCStr, *pCStr1, CStr;
  POSITION pos, pos1;

  // do we have a title ?
  if (Title.IsSetTitle())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_TITLE)) return FALSE;

    // write the title itself
    CStr = KEY_TITLE;
    CStr = CStr + "           = ";
    CStr = CStr + Title.GetTitle();
    if (!WriteCfgLine(CStr)) return FALSE;
  }

  // do we have title lines ?
  if ((Title.GetTitleLines())->GetCount() > 0)
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_TITLELINE)) return FALSE;

    // write the title lines
    CStr = KEY_TITLELINE;
    CStr = CStr + "       = ";
    // iterate through all lines
    pos1 = (Title.GetTitleVoices())->GetHeadPosition();
    for (pos = (Title.GetTitleLines())->GetHeadPosition(); pos != NULL; )
    {
      // get the title line
      pCStr = (CString *)(Title.GetTitleLines())->GetNext(pos);
      // get the optional voice file
      if (pos1 != NULL)
      {
        pCStr1 = (CString *)(Title.GetTitleVoices())->GetNext(pos1);
        // do we have a voice file ?
        if (pCStr1->IsEmpty())
        {
          // no, just write the title line
          if (!WriteCfgLine(CStr + *pCStr)) return FALSE;
        } else
        {
          // yes, write the title line and the voice file
          if (!WriteCfgLine(CStr + "[" + *pCStr1 +"] " + *pCStr)) return FALSE;
        }
      }
    }
  }

  return TRUE;
}

//-----------------------------------
// write the epiloglines
//-----------------------------------
BOOL CConfiguration::WriteEpilogLines()
{
  CString  *pCStr, *pCStr1, CStr;
  POSITION pos, pos1;

  // do we have epilog lines ?
  if ((Epilog.GetEpilogLines())->GetCount() > 0)
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_EPILOGLINE)) return FALSE;

    // write the epilog lines
    CStr = KEY_EPILOGLINE;
    CStr = CStr + "      = ";
    // iterate through all lines
    pos1 = (Epilog.GetEpilogVoices())->GetHeadPosition();
    for (pos = (Epilog.GetEpilogLines())->GetHeadPosition(); pos != NULL; )
    {
      // get the epilog line
      pCStr = (CString *)(Epilog.GetEpilogLines())->GetNext(pos);
      // get the optional voice file
      if (pos1 != NULL)
      {
        pCStr1 = (CString *)(Epilog.GetEpilogVoices())->GetNext(pos1);
        // do we have a voice file ?
        if (pCStr1->IsEmpty())
        {
          // no, just write the epilog line
          if (!WriteCfgLine(CStr + *pCStr)) return FALSE;
        } else
        {
          // yes, write the epilog line and the voice file
          if (!WriteCfgLine(CStr + "[" + *pCStr1 +"] " + *pCStr)) return FALSE;
        }
      }
    }
  }

  return TRUE;
}

//-----------------------------------
// write the player and his escort
//-----------------------------------
BOOL CConfiguration::WritePlayerAndEscort()
{
  CString   CStr, CStr1;
  CShipData *pShip;
  POSITION  pos;

  // do we have a player ? (we should have one)
  if (PlayerShip.IsSetShipClass())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_PLAYER)) return FALSE;

    // write the player itself
    CStr = KEY_PLAYER;
    CStr = CStr + "          = ";
    OutputShipClassAndName(&CStr1, &PlayerShip);

    // write it
    if (!WriteCfgLine(CStr + CStr1)) return FALSE;
  }

  // do we have an escort ?
  if (Escort.GetCount() > 0)
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_ESCORT)) return FALSE;

    // write the escort
    CStr = KEY_ESCORT;
    CStr = CStr + "          = ";
    // iterate through all escorts
    for (pos = Escort.GetHeadPosition(); pos != NULL; )
    {
      // get ship class and name
      pShip = (CShipData *)Escort.GetNext(pos);
      OutputShipClassAndName(&CStr1, pShip);

      // write it
      if (!WriteCfgLine(CStr + CStr1)) return FALSE;
    }
  }

  return TRUE;
}

//-----------------------------------
// write general parameters
//-----------------------------------
BOOL CConfiguration::WriteGeneral()
{
  CString   CStr;

  // warp out mode set
  if (GeneralMode.IsSetWarpOutMode())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_WARPOUT)) return FALSE;

    // write the mode
    CStr = KEY_WARPOUT;
    CStr = CStr + "         = ";
    if (GeneralMode.GetWarpOutMode() == Anytime)
    {
      CStr = CStr + SUBKEY_ANYTIME;
    } else
    if (GeneralMode.GetWarpOutMode() == WhenCleared)
    {
      CStr = CStr + SUBKEY_CLEARED;
    } else
    {
      CStr = CStr + SUBKEY_SEQUENTIAL;
    }

    // write it
    if (!WriteCfgLine(CStr)) return FALSE;
  }
    
  // collision mode set
  if (GeneralMode.IsSetCollisionMode())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_COLLISION)) return FALSE;

    // write the mode
    CStr = KEY_COLLISION;
    CStr = CStr + "       = ";
    if (GeneralMode.GetCollisionMode() == CollisionsOn)
    {
      CStr = CStr + SUBKEY_ON;
    } else
    if (GeneralMode.GetCollisionMode() == CollisionsOff)
    {
      CStr = CStr + SUBKEY_OFF;
    } else
    {
      CStr = CStr + SUBKEY_OFFALL;
    }

    // write it
    if (!WriteCfgLine(CStr)) return FALSE;
  }

  // ship change mode set
  if (GeneralMode.IsSetShipChangeMode())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_SHIPCHANGE)) return FALSE;

    // write the mode
    CStr = KEY_SHIPCHANGE;
    CStr = CStr + "      = ";
    if (GeneralMode.GetShipChangeMode() == NoShipChange)
    {
      CStr = CStr + SUBKEY_OFF;
    } else
    if (GeneralMode.GetShipChangeMode() == ManualShipChange)
    {
      CStr = CStr + SUBKEY_ON;
    } else
    {
      CStr = CStr + SUBKEY_AUTO;
    }

    // write it
    if (!WriteCfgLine(CStr)) return FALSE;
  }

  // perfect cloak mode set
  if (GeneralMode.IsSetPerfectCloakMode())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_PERFECTCLOAK)) return FALSE;

    // write the mode
    CStr = KEY_PERFECTCLOAK;
    CStr = CStr + "    = ";
    if (GeneralMode.GetPerfectCloakMode() == PerfectCloakOn)
    {
      CStr = CStr + SUBKEY_ON;
    } else
    {
      CStr = CStr + SUBKEY_OFF;
    }

    // write it
    if (!WriteCfgLine(CStr)) return FALSE;
  }

  return TRUE;
}

//-----------------------------------
// write self destruct
//-----------------------------------
BOOL CConfiguration::WriteSelfDestruct()
{
  CString   CStr;
  char      str[MAX_QB2_LINELENGTH + 1];

  // self destruct defined
  if (SelfDestruct.IsSetMode())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_SELFDESTRUCT)) return FALSE;

    // write the self destruct mode
    CStr = KEY_SELFDESTRUCT;
    CStr = CStr + "    = ";
    if (SelfDestruct.GetMode() == StandardDestruct)
    {
      CStr = CStr + SUBKEY_STANDARD;
    } else
    if (SelfDestruct.GetMode() == OverkillDestruct)
    {
      CStr = CStr + SUBKEY_OVERKILL;
      sprintf(str," %li",SelfDestruct.GetOverkillDistance());
      CStr = CStr + str;
    } else
    {
      CStr = CStr + SUBKEY_SELOVERKILL;
      sprintf(str," %li",SelfDestruct.GetOverkillDistance());
      CStr = CStr + str;
    }
 
    // add a code if code is set
    if (!(SelfDestruct.GetCode().IsEmpty()))
    {
      CStr = CStr + " = ";
      CStr = CStr + SelfDestruct.GetCode();
    }

    // write it
    if (!WriteCfgLine(CStr)) return FALSE;
  }
    

  return TRUE;
}


//-----------------------------------
// write the system name
//-----------------------------------
BOOL CConfiguration::WriteSystemName(CSystemData *pSystem)
{
  CString   CStr;

  // do we have a system name ? (we should have anyway)
  if (pSystem->IsSetSystemName())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_SYSTEM)) return FALSE;

    // write the system name
    CStr = KEY_SYSTEM;
    CStr = CStr + "          = ";
    CStr = CStr + pSystem->GetSystemName();

    // write it
    if (!WriteCfgLine(CStr)) return FALSE;
  }

  return TRUE;
}

//-----------------------------------
// write the friends in the given system
//-----------------------------------
BOOL CConfiguration::WriteFriends(CSystemData *pSystem)
{
  CString   CStr, CStr1;
  CShipData *pShip;
  POSITION  pos;

  // do we have friends ?
  if ((pSystem->GetFriends())->GetCount() > 0)
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_FRIENDS)) return FALSE;

    // write the friends
    CStr = KEY_FRIEND;
    CStr = CStr + "          = ";
    // iterate through all friends
    for (pos = (pSystem->GetFriends())->GetHeadPosition(); pos != NULL; )
    {
      // get ship class and name
      pShip = (CShipData *)(pSystem->GetFriends())->GetNext(pos);
      OutputShipClassAndName(&CStr1, pShip);

      // write it
      if (!WriteCfgLine(CStr + CStr1)) return FALSE;
    }
  }

  return TRUE;
}

//-----------------------------------
// write the enemies in the given system
//-----------------------------------
BOOL CConfiguration::WriteEnemies(CSystemData *pSystem)
{
  CString   CStr, CStr1;
  CShipData *pShip;
  POSITION  pos;

  // do we have friends ?
  if ((pSystem->GetEnemies())->GetCount() > 0)
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_ENEMIES)) return FALSE;

    // write the enemies
    CStr = KEY_ENEMY;
    CStr = CStr + "           = ";
    // iterate through all enemies
    for (pos = (pSystem->GetEnemies())->GetHeadPosition(); pos != NULL; )
    {
      // get ship class and name
      pShip = (CShipData *)(pSystem->GetEnemies())->GetNext(pos);
      OutputShipClassAndName(&CStr1, pShip);

      // write it
      if (!WriteCfgLine(CStr + CStr1)) return FALSE;
    }
  }

  return TRUE;
}

//-----------------------------------
// write the neutrals in the given system
//-----------------------------------
BOOL CConfiguration::WriteNeutrals(CSystemData *pSystem)
{
  CString   CStr, CStr1;
  CShipData *pShip;
  POSITION  pos;

  // do we have neutrals ?
  if ((pSystem->GetNeutrals())->GetCount() > 0)
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_NEUTRALS)) return FALSE;

    // write the neutrals
    CStr = KEY_NEUTRAL;
    CStr = CStr + "         = ";
    // iterate through all neutrals
    for (pos = (pSystem->GetNeutrals())->GetHeadPosition(); pos != NULL; )
    {
      // get ship class and name
      pShip = (CShipData *)(pSystem->GetNeutrals())->GetNext(pos);
      OutputShipClassAndName(&CStr1, pShip);

      // write it
      if (!WriteCfgLine(CStr + CStr1)) return FALSE;
    }
  }

  return TRUE;
}


//-----------------------------------
// write the hulks in the given system
//-----------------------------------
BOOL CConfiguration::WriteHulks(CSystemData *pSystem)
{
  CString   CStr, CStr1;
  CShipData *pShip;
  POSITION  pos;

  // do we have hulks ?
  if ((pSystem->GetHulks())->GetCount() > 0)
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_HULKS)) return FALSE;

    // write the hulks
    CStr = KEY_HULK;
    CStr = CStr + "            = ";
    // iterate through all hulks
    for (pos = (pSystem->GetHulks())->GetHeadPosition(); pos != NULL; )
    {
      // get ship class and name
      pShip = (CShipData *)(pSystem->GetHulks())->GetNext(pos);
      OutputShipClassAndName(&CStr1, pShip);

      // write it
      if (!WriteCfgLine(CStr + CStr1)) return FALSE;
    }
  }

  return TRUE;
}

//-----------------------------------
// write the jump gate in the given system
//-----------------------------------
BOOL CConfiguration::WriteJumpGate(CSystemData *pSystem)
{
  CString   CStr, CStr1;

  // do we have a jump gate ? 
  if (pSystem->GetJumpGate()->IsSetShipClass())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_JUMPGATE)) return FALSE;

    // write the jump gate
    CStr = KEY_JUMPGATE;
    CStr = CStr + "        = ";
    OutputShipClassAndName(&CStr1, pSystem->GetJumpGate());

    // write it
    if (!WriteCfgLine(CStr + CStr1)) return FALSE;
  }

  return TRUE;
}


//-----------------------------------
// write the jump time in the given system
//-----------------------------------
BOOL CConfiguration::WriteJumpTime(CSystemData *pSystem)
{
  CString  CStr;
  char     str[MAX_QB2_LINELENGTH + 1];

  // do we have a jump time ?
  if (pSystem->IsSetJumpTime())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_JUMPTIME)) return FALSE;

    // set the key
    CStr = KEY_JUMPTIME;
    sprintf(str,"        = %li",pSystem->GetJumpTime());
    if (!WriteCfgLine(CStr + str)) return FALSE;
  }

  return TRUE;

}

//-----------------------------------
// write the jump destination in the given system
//-----------------------------------
BOOL CConfiguration::WriteJumpDestination(CSystemData *pSystem)
{
  CString  CStr;

  // do we have a jump destination ?
  if (pSystem->IsSetJumpDestination())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_JUMPDEST)) return FALSE;

    // set the key
    CStr = KEY_JUMPDEST;
    CStr = CStr + " = ";
    CStr = CStr + pSystem->GetJumpDestination();
    if (!WriteCfgLine(CStr)) return FALSE;
  }

  return TRUE;

}

//-----------------------------------
// write the jump options in the given system
//-----------------------------------
BOOL CConfiguration::WriteJumpOptions(CSystemData *pSystem)
{
  CString  CStr;
  char     str[MAX_QB2_LINELENGTH + 1];

  // do we have jump options ?
  if (pSystem->GetJumpOptions()->IsSetJumpOnly()           ||
      pSystem->GetJumpOptions()->IsSetClearJump()          ||
      pSystem->GetJumpOptions()->IsSetNoSpin()             ||
      pSystem->GetJumpOptions()->IsSetShowActivation()     ||
      pSystem->GetJumpOptions()->IsSetActOnHail()          ||
      pSystem->GetJumpOptions()->IsSetGateHidden()         ||
      pSystem->GetJumpOptions()->IsSetActivationDistance())

  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_JUMPOPTIONS)) return FALSE;

    // set the key
    CStr = KEY_JUMPOPTION;
    CStr = CStr + "      = ";

    // option gate hidden set ?
    if (pSystem->GetJumpOptions()->IsSetGateHidden())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_GATEHIDDEN)) return FALSE;
    }

    // option jump only set ?
    if (pSystem->GetJumpOptions()->IsSetJumpOnly())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_JUMPONLY)) return FALSE;
    }

    // option clear on jump only set ?
    if (pSystem->GetJumpOptions()->IsSetClearJump())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_CLEARJUMP)) return FALSE;
    }

    // option no spin set ?
    if (pSystem->GetJumpOptions()->IsSetNoSpin())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_NOSPIN)) return FALSE;
    }

    // option show activation set ?
    if (pSystem->GetJumpOptions()->IsSetShowActivation())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_SHOWACT)) return FALSE;
    }

    // option activate on hail set ?
    if (pSystem->GetJumpOptions()->IsSetActOnHail())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_ACTONHAIL)) return FALSE;
    }

    // option activation distance set ?
    if (pSystem->GetJumpOptions()->IsSetActivationDistance())
    {
      // write it
      sprintf(str,"%s %li",SUBKEY_ACTDISTANCE, pSystem->GetJumpOptions()->GetActivationDistance());
      if (!WriteCfgLine(CStr + str)) return FALSE;
    }
  }

  return TRUE;
}


//-----------------------------------
// write the positions in the given system
//-----------------------------------
BOOL CConfiguration::WritePositions(CSystemData *pSystem)
{
  CString   CStr;
  char      str[MAX_QB2_LINELENGTH + 1];

  // do we have positions ?
  if (pSystem->IsSetPlayerPos()   || pSystem->IsSetPlayerPosPlanet()   ||
      pSystem->IsSetEnemyPos()    || pSystem->IsSetEnemyPosPlanet()    ||
      pSystem->IsSetFriendPos()   || pSystem->IsSetFriendPosPlanet()   ||
      pSystem->IsSetNeutralPos()  || pSystem->IsSetNeutralPosPlanet()  ||
      pSystem->IsSetHulkPos()     || pSystem->IsSetHulkPosPlanet()     ||
      pSystem->IsSetJumpGatePos() || pSystem->IsSetJumpGatePosPlanet())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_POSITION)) return FALSE;

    // write the player position
    if (pSystem->IsSetPlayerPos())
    {
      CStr = KEY_PLAYERPOS;
      sprintf(str,"  = %li %li %li",pSystem->GetPlayerPosX(), pSystem->GetPlayerPosY(), pSystem->GetPlayerPosZ());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    } else
    if (pSystem->IsSetPlayerPosPlanet())
    {
      CStr = KEY_PLAYERPOS;
      sprintf(str,"  = %s %s",SUBKEY_PLANET, pSystem->GetPlayerPosPlanet());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    }

    // write the friend position
    if (pSystem->IsSetFriendPos())
    {
      CStr = KEY_FRIENDPOS;
      sprintf(str,"  = %li %li %li",pSystem->GetFriendPosX(), pSystem->GetFriendPosY(), pSystem->GetFriendPosZ());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    } else
    if (pSystem->IsSetFriendPosPlanet())
    {
      CStr = KEY_FRIENDPOS;
      sprintf(str,"  = %s %s",SUBKEY_PLANET, pSystem->GetFriendPosPlanet());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    }

    // write the enemy position
    if (pSystem->IsSetEnemyPos())
    {
      CStr = KEY_ENEMYPOS;
      sprintf(str,"   = %li %li %li",pSystem->GetEnemyPosX(), pSystem->GetEnemyPosY(), pSystem->GetEnemyPosZ());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    } else
    if (pSystem->IsSetEnemyPosPlanet())
    {
      CStr = KEY_ENEMYPOS;
      sprintf(str,"   = %s %s",SUBKEY_PLANET, pSystem->GetEnemyPosPlanet());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    }

    // write the neutral position
    if (pSystem->IsSetNeutralPos())
    {
      CStr = KEY_NEUTRALPOS;
      sprintf(str," = %li %li %li",pSystem->GetNeutralPosX(), pSystem->GetNeutralPosY(), pSystem->GetNeutralPosZ());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    } else
    if (pSystem->IsSetNeutralPosPlanet())
    {
      CStr = KEY_NEUTRALPOS;
      sprintf(str," = %s %s",SUBKEY_PLANET, pSystem->GetNeutralPosPlanet());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    }

    // write the hulk position
    if (pSystem->IsSetHulkPos())
    {
      CStr = KEY_HULKPOS;
      sprintf(str,"    = %li %li %li",pSystem->GetHulkPosX(), pSystem->GetHulkPosY(), pSystem->GetHulkPosZ());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    } else
    if (pSystem->IsSetHulkPosPlanet())
    {
      CStr = KEY_HULKPOS;
      sprintf(str,"    = %s %s",SUBKEY_PLANET, pSystem->GetHulkPosPlanet());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    }

    // write the jump gate position
    if (pSystem->IsSetJumpGatePos())
    {
      CStr = KEY_JUMPPOS;
      sprintf(str,"    = %li %li %li",pSystem->GetJumpGatePosX(), pSystem->GetJumpGatePosY(), pSystem->GetJumpGatePosZ());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    } else
    if (pSystem->IsSetJumpGatePosPlanet())
    {
      CStr = KEY_JUMPPOS;
      sprintf(str,"    = %s %s",SUBKEY_PLANET, pSystem->GetJumpGatePosPlanet());
      CStr = CStr + str;
      if (!WriteCfgLine(CStr)) return FALSE;
    }

  }

  return TRUE;

}

//-----------------------------------
// write the dialogs in the given system
//-----------------------------------
BOOL CConfiguration::WriteDialogs(CSystemData *pSystem)
{
  CString     CStr, CStr1, Rem;
  CDialogList *pDlgList;
  CDialogData *pDialog;
  POSITION    pos;
  unsigned    i;

  // loop through all dialog types
  for (i=0; i <= 12; i++)
  {
    // select dialog type
    switch (i)
    {
      // starting dialog
      case 0: pDlgList = pSystem->GetStartingDialogs();
              CStr = KEY_STARTDIALOG;
              CStr = CStr + "  = ";
              Rem  = REM_STARTDIALOG;
              break;

      // ending dialog
      case 1: pDlgList = pSystem->GetEndingDialogs();
              CStr = KEY_ENDDIALOG;
              CStr = CStr + "    = ";
              Rem  = REM_ENDDIALOG;
              break;

      // friend warp in dialog
      case 2: pDlgList = pSystem->GetFriendWarpInDialogs();
              CStr = KEY_FRIENDWARPINDIALOG;
              CStr = CStr + "   = ";
              Rem  = REM_FRIENDWARPINDIALOG;
              break;

      // enemy warp in dialog
      case 3: pDlgList = pSystem->GetEnemyWarpInDialogs();
              CStr = KEY_ENEMYWARPINDIALOG;
              CStr = CStr + "    = ";
              Rem  = REM_ENEMYWARPINDIALOG;
              break;

      // neutral warp in dialog
      case 4: pDlgList = pSystem->GetNeutralWarpInDialogs();
              CStr = KEY_NEUTRALWARPINDIALOG;
              CStr = CStr + "  = ";
              Rem  = REM_NEUTRALWARPINDIALOG;
              break;

      // friend warp out dialog
      case 5: pDlgList = pSystem->GetFriendWarpOutDialogs();
              CStr = KEY_FRIENDWARPOUTDIALOG;
              CStr = CStr + "  = ";
              Rem  = REM_FRIENDWARPOUTDIALOG;
              break;

      // enemy warp out dialog
      case 6: pDlgList = pSystem->GetEnemyWarpOutDialogs();
              CStr = KEY_ENEMYWARPOUTDIALOG;
              CStr = CStr + "   = ";
              Rem  = REM_ENEMYWARPOUTDIALOG;
              break;

      // neutral warp out dialog
      case 7: pDlgList = pSystem->GetNeutralWarpOutDialogs();
              CStr = KEY_NEUTRALWARPOUTDIALOG;
              CStr = CStr + " = ";
              Rem  = REM_NEUTRALWARPOUTDIALOG;
              break;

      // friend join escort dialog
      case 8: pDlgList = pSystem->GetFriendJoinEscortDialogs();
              CStr = KEY_JOINESCORTDIALOG;
              CStr = CStr + " = ";
              Rem  = REM_JOINESCORTDIALOG;
              break;

      // neutral to friend dialog
      case 9: pDlgList = pSystem->GetNeutralToFriendDialogs();
              CStr = KEY_NEUTRALTOFRIENDDIALOG;
              CStr = CStr + " = ";
              Rem  = REM_NEUTRALTOFRIENDDIALOG;
              break;

      // neutral to enemy dialog
      case 10: pDlgList = pSystem->GetNeutralToEnemyDialogs();
               CStr = KEY_NEUTRALTOENEMYDIALOG;
               CStr = CStr + "  = ";
               Rem  = REM_NEUTRALTOENEMYDIALOG;
               break;

      // jump gate appears dialog
      case 11: pDlgList = pSystem->GetJumpGateAppearsDialogs();
               CStr = KEY_JUMPAPPEARSDIALOG;
               CStr = CStr + " = ";
               Rem  = REM_JUMPAPPEARSDIALOG;
               break;

      // jump dialog
      case 12: pDlgList = pSystem->GetJumpDialogs();
               CStr = KEY_JUMPDIALOG;
               CStr = CStr + "      = ";
               Rem  = REM_JUMPDIALOG;
               break;

      default: break;
    }

    // do we have dialogs of the current type?
    if (pDlgList->GetCount() > 0)
    {
      // write a blank line
      if (!WriteCfgLine("")) return FALSE;

      // write a remark
      if (!WriteCfgLine(Rem)) return FALSE;

      // iterate through all starting dialogs
      for (pos = pDlgList->GetHeadPosition(); pos != NULL; )
      {
        // get the dialog
        pDialog = (CDialogData *)pDlgList->GetNext(pos);
        CStr1 = "";
        // do we have a voice file ?
        if (!pDialog->GetVoiceFile().IsEmpty()) CStr1 = CStr1 + "[" + pDialog->GetVoiceFile() + "] ";
        CStr1 = CStr1 + pDialog->GetSpeaker() + ": ";
        CStr1 = CStr1 + pDialog->GetDialogLine();

        // write it
        if (!WriteCfgLine(CStr + CStr1)) return FALSE;
      }
    }
  }

  return TRUE;
}


//-----------------------------------
// write the warp disables in the given system
//-----------------------------------
BOOL CConfiguration::WriteWarpDisable(CSystemData *pSystem)
{
  CString  CStr;

  // do we have warp disables ?
  if (pSystem->IsSetPlayerWarpDisable()  || pSystem->IsSetFriendWarpDisable()  ||
      pSystem->IsSetEnemyWarpDisable()   || pSystem->IsSetNeutralWarpDisable())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_WARPDISABLE)) return FALSE;

    // set the key
    CStr = KEY_NOWARP;
    CStr = CStr + "     = ";

    // do we have a player warp disable
    if (pSystem->IsSetPlayerWarpDisable())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_PLAYER)) return FALSE;
    }

    // do we have a friend warp disable
    if (pSystem->IsSetFriendWarpDisable())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_FRIEND)) return FALSE;
    }

    // do we have an enemy warp disable
    if (pSystem->IsSetEnemyWarpDisable())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_ENEMY)) return FALSE;
    }

    // do we have a neutral warp disable
    if (pSystem->IsSetNeutralWarpDisable())
    {
      // write it
      if (!WriteCfgLine(CStr + SUBKEY_NEUTRAL)) return FALSE;
    }
  }

  return TRUE;
}

//-----------------------------------
// write the stardate in the given system
//-----------------------------------
BOOL CConfiguration::WriteStardate(CSystemData *pSystem)
{
  CString  CStr;
  char     str[MAX_QB2_LINELENGTH + 1];

  // do we have a stardate ?
  if (pSystem->IsSetStardate())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_STARDATE)) return FALSE;

    // set the key
    CStr = KEY_STARDATE;
    sprintf(str,"        = %li",pSystem->GetStardate());
    if (!WriteCfgLine(CStr + str)) return FALSE;
  }

  return TRUE;

}

//-----------------------------------
// write the neutral behavior in the given system
//-----------------------------------
BOOL CConfiguration::WriteNeutralBehavior(CSystemData *pSystem)
{
  CString  CStr;
  char     str[MAX_QB2_LINELENGTH + 1];

  // do we have a neutral behavior ?
  if (pSystem->IsSetNeutralBehavior())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_NEUTRALBEHAV)) return FALSE;

    // set the key
    CStr = KEY_NEUTRALBEHAVIOR;
    CStr = CStr + " = ";

    // write the behavior
    if (pSystem->GetNeutralBehavior() == EnemyWhenAttacked)
    {
      CStr = CStr + SUBKEY_ENEMYWHENATTACK;
    } else
    if (pSystem->GetNeutralBehavior() == EnemyAfter)
    {
      CStr = CStr + SUBKEY_ENEMYAFTER;
      sprintf(str," %li",pSystem->GetNeutralBehaviorTime());
      CStr = CStr + str;   
    } else
    if (pSystem->GetNeutralBehavior() == FriendAfter)
    {
      CStr = CStr + SUBKEY_FRIENDAFTER;
      sprintf(str," %li",pSystem->GetNeutralBehaviorTime());
      CStr = CStr + str;   
    } else
    {
      CStr = CStr + SUBKEY_RANDOM;
    }
    if (!WriteCfgLine(CStr)) return FALSE;
  }
  return TRUE;
}

//-----------------------------------
// write the actions in the given system
//-----------------------------------
BOOL CConfiguration::WriteActions(CSystemData *pSystem)
{
  CString  CStr, CStr1;
  char     str[MAX_QB2_LINELENGTH + 1];

  // do we have friend actions ?
  if (pSystem->IsSetFriendWarpInTime() || pSystem->IsSetFriendWarpOutTime() || pSystem->IsSetFriendJoinEscortTime())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_FRIENDACTION)) return FALSE;

    // set the key
    CStr = KEY_FRIENDACTION;
    CStr = CStr + "    = ";

    // which friend actions ?
    // warp in ?
    if (pSystem->IsSetFriendWarpInTime())
    {
      CStr1 = CStr + SUBKEY_WARPIN + " ";
      // get the time
      sprintf(str,"%li",pSystem->GetFriendWarpInTime());
      CStr1 = CStr1 + str;   
      if (!WriteCfgLine(CStr1)) return FALSE;
    }

    // warp out ?
    if (pSystem->IsSetFriendWarpOutTime())
    {
      CStr1 = CStr + SUBKEY_WARPOUT + " ";
      // get the time
      if (pSystem->GetFriendWarpOutTime() >= 0)
      {
        // time was specified
        sprintf(str,"%li",pSystem->GetFriendWarpOutTime());
        CStr1 = CStr1 + str;   
      } else
      {
        // cleared was specified
        CStr1 = CStr1 + SUBKEY_TCLEARED;   
      }
      if (!WriteCfgLine(CStr1)) return FALSE;
    }

    // join escort ?
    if (pSystem->IsSetFriendJoinEscortTime())
    {
      CStr1 = CStr + SUBKEY_JOINESCORT + " ";
      // get the time
      if (pSystem->GetFriendJoinEscortTime() >= 0)
      {
        // time was specified
        sprintf(str,"%li",pSystem->GetFriendJoinEscortTime());
        CStr1 = CStr1 + str;   
      } else
      {
        // cleared was specified
        CStr1 = CStr1 + SUBKEY_TCLEARED;   
      }
      if (!WriteCfgLine(CStr1)) return FALSE;
    }
  }

  // do we have enemy actions ?
  if (pSystem->IsSetEnemyWarpInTime() || pSystem->IsSetEnemyWarpOutTime())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_ENEMYACTION)) return FALSE;

    // set the key
    CStr = KEY_ENEMYACTION;
    CStr = CStr + "     = ";

    // which enemy actions ?
    // warp in ?
    if (pSystem->IsSetEnemyWarpInTime())
    {
      CStr1 = CStr + SUBKEY_WARPIN + " ";
      // get the time
      sprintf(str,"%li",pSystem->GetEnemyWarpInTime());
      CStr1 = CStr1 + str;   
      if (!WriteCfgLine(CStr1)) return FALSE;
    }

    // warp out ?
    if (pSystem->IsSetEnemyWarpOutTime())
    {
      CStr1 = CStr + SUBKEY_WARPOUT + " ";
      // get the time
      if (pSystem->GetEnemyWarpOutTime() >= 0)
      {
        // time was specified
        sprintf(str,"%li",pSystem->GetEnemyWarpOutTime());
        CStr1 = CStr1 + str;   
      } else
      {
        // cleared was specified
        CStr1 = CStr1 + SUBKEY_TCLEARED;   
      }
      if (!WriteCfgLine(CStr1)) return FALSE;
    }
  }

  // do we have neutral actions ?
  if (pSystem->IsSetNeutralWarpInTime() || pSystem->IsSetNeutralWarpOutTime())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_NEUTRALACTION)) return FALSE;

    // set the key
    CStr = KEY_NEUTRALACTION;
    CStr = CStr + "   = ";

    // which neutral actions ?
    // warp in ?
    if (pSystem->IsSetNeutralWarpInTime())
    {
      CStr1 = CStr + SUBKEY_WARPIN + " ";
      // get the time
      sprintf(str,"%li",pSystem->GetNeutralWarpInTime());
      CStr1 = CStr1 + str;   
      if (!WriteCfgLine(CStr1)) return FALSE;
    }

    // warp out ?
    if (pSystem->IsSetNeutralWarpOutTime())
    {
      CStr1 = CStr + SUBKEY_WARPOUT + " ";
      // get the time
      if (pSystem->GetNeutralWarpOutTime() >= 0)
      {
        // time was specified
        sprintf(str,"%li",pSystem->GetNeutralWarpOutTime());
        CStr1 = CStr1 + str;   
      } else
      {
        // cleared was specified
        CStr1 = CStr1 + SUBKEY_TCLEARED;   
      }
      if (!WriteCfgLine(CStr1)) return FALSE;
    }
  }


  return TRUE;
}

//-----------------------------------
// write the battle mode in the given system
//-----------------------------------
BOOL CConfiguration::WriteBattleMode(CSystemData *pSystem)
{
  CString  CStr;

  // do we have a battle mode ?
  if (pSystem->IsSetBattleMode())
  {
    // write a blank line
    if (!WriteCfgLine("")) return FALSE;

    // write a remark
    if (!WriteCfgLine(REM_BATTLEMODE)) return FALSE;

    // set the key
    CStr = KEY_BATTLEMODE;
    CStr = CStr + "      = ";

    // write the battle mode
    if (pSystem->GetBattleMode() == StatisticBattle)
    {
      CStr = CStr + SUBKEY_STATBATTLE;
    } else
    {
      CStr = CStr + SUBKEY_AIBATTLE;
    }
    if (!WriteCfgLine(CStr)) return FALSE;
  }
  return TRUE;
}


// ----------  the available ships class --------------------------------------

//-----------------------------------
// the constructor
//-----------------------------------
CAvailableShips::CAvailableShips()
{
}

//-----------------------------------
// the destructor
//-----------------------------------
CAvailableShips::~CAvailableShips()
{
}

//-----------------------------------
// get the list of available ships
//-----------------------------------
BOOL CAvailableShips::Get()
{
  CString      strFileName;
  CShipData    *pShip;
  CErrorDialog ErrorDialog(NULL);
  char         line[LINE_LENGTH], line1[LINE_LENGTH];

  // check if we have already done this
  if (AvailableShipList.GetCount() > 0) return TRUE;

  // get the name of the available ship file
  strFileName = strPathToKA + QB2_PATH + QB2_SHIPFILE;

  // open the ship file
  if (!AvailableShipFile.Open(strFileName))
  {
    // an error
    ErrorDialog.Report(ERR_AVSHIPOPEN, ERR_HEADER);
    return FALSE;
  }

  // read all lines
  while (AvailableShipFile.ReadLine(line))
  {

    // look for the = sign
    Tokenize(line1, line, '=');
    if (strlen(line) > strlen(line1))
    {
      // found, isolate ship class
      GetSubString(line, line, strlen(line1) + 1, strlen(line) - 1);
      StripString(line,line);
      // isolate the name
      StripString(line1, line1);
      pShip = new CShipData;
      if (pShip == NULL)
      {
        // an error
       ErrorDialog.Report(ERR_OUTOFMEMORY, ERR_HEADER);
       return FALSE;
      }
      pShip->SetShipClass(line);
      pShip->SetShipName(line1);
      // add it to the list
      AvailableShipList.AddTail(pShip);
    }
  }

  // close the file
  AvailableShipFile.Close();

  return TRUE;

}

//-----------------------------------
// find a ship class
//-----------------------------------
BOOL CAvailableShips::FindShipClass(LONG Index, CString *Class)
{
  POSITION  pos;
  CShipData *pShip;

  // search the index
  pos = AvailableShipList.FindIndex(Index);
  if (pos == NULL) return FALSE;

  // get the class
  pShip = (CShipData *)AvailableShipList.GetAt(pos);
  *Class = pShip->GetShipClass();

  return TRUE;
}

//-----------------------------------
// find a ship
//-----------------------------------
CShipData *CAvailableShips::FindShip(CString Class)
{
  POSITION  pos;
  CShipData *pShip;

  // loop through all ship classes
  for (pos = AvailableShipList.GetHeadPosition(); pos != NULL; )
  {
    // compare ship class
    pShip = (CShipData *)AvailableShipList.GetNext(pos);
    if (pShip->GetShipClass().CompareNoCase(Class) == 0) return pShip;
  }

  return NULL;

}

//-----------------------------------
// fill a listbox with the ships
//-----------------------------------
void CAvailableShips::FillListBox(CModalDialog *pDialog, LONG ListBoxId)
{
  POSITION  pos;
  CShipData *pShip;
  CString   CStr;

  // copy the ship list to the list box
  for (pos = AvailableShipList.GetHeadPosition(); pos != NULL; )
  {
    // get ship class and name
    pShip = (CShipData *)AvailableShipList.GetNext(pos);
    AnotherFormatShipClassAndName(&CStr, pShip);

    // put it in the box
    pDialog->SendDlgItemMessage(ListBoxId, LB_ADDSTRING,0,(LONG)(const char *)CStr);
  }
}


// ----------  the system planet data class, contains system name and planet names for a specific system ----

//-----------------------------------
// the constructor
//-----------------------------------
CSystemPlanetData::CSystemPlanetData()
{
  // init all protected variables
  m_SystemName = "";
}


// ----------  the system planet list class, contains system name and planet names for some specific systems ----

//-----------------------------------
// find a system
//-----------------------------------
CSystemPlanetData *CSystemPlanetList::FindSystem(const char *strSystemName)
{
  POSITION          pos;
  CSystemPlanetData *pSystemPlanet;

  // iterate the list
  for (pos = GetHeadPosition(); pos != NULL; )
  {
    // get the description
    pSystemPlanet = (CSystemPlanetData *)GetNext(pos);
    if ((pSystemPlanet->GetSystemName()).CompareNoCase(strSystemName) == 0) return pSystemPlanet;
  }

  return NULL;
}


// ----------  the available systems and planets class --------------------------------------

//-----------------------------------
// the constructor
//-----------------------------------
CAvailableSystemsAndPlanets::CAvailableSystemsAndPlanets()
{
}

//-----------------------------------
// the destructor
//-----------------------------------
CAvailableSystemsAndPlanets::~CAvailableSystemsAndPlanets()
{
}

//-----------------------------------
// get the list of available systems and planets
//-----------------------------------
BOOL CAvailableSystemsAndPlanets::Get()
{
  CString            strFileName, *pPlanet;
  CSystemPlanetData  *pSystemPlanet;
  CErrorDialog       ErrorDialog(NULL);
  char               line[LINE_LENGTH], line1[LINE_LENGTH];

  // check if we have already done this
  if (AvailableSystemsPlanetsList.GetCount() > 0) return TRUE;

  // get the name of the available systems and planets file
  strFileName = strPathToKA + QB2_PATH + QB2_SYSTEMPLANETFILE;

  // open the ship file
  if (!AvailableSystemsPlanetsFile.Open(strFileName))
  {
    // an error
    ErrorDialog.Report(ERR_AVSYSTEMOPEN, ERR_HEADER);
    return FALSE;
  }

  // read all lines
  while (AvailableSystemsPlanetsFile.ReadLine(line))
  {

    // parse the line
    // first is the system name
    Tokenize(line1, line, '=');
    // remove it from line
    GetSubString(line, line, strlen(line1) + 1, strlen(line) - 1);

    // check the system name
    StripString(line1, line1);
    if (line1[0] !='\0')
    {
      // create a new system planet data object
      pSystemPlanet = new CSystemPlanetData;
      if (pSystemPlanet == NULL)
      {
        // an error
       ErrorDialog.Report(ERR_OUTOFMEMORY, ERR_HEADER);
       return FALSE;
      }
      // set the system name
      pSystemPlanet->SetSystemName(line1);
      // add it to the system planet list
      AvailableSystemsPlanetsList.AddTail(pSystemPlanet);

      // look for planets
      do
      {
        // get a planet name
        Tokenize(line1, line, '=');
        // remove it from line
        GetSubString(line, line, strlen(line1) + 1, strlen(line) - 1);
        // check the planet name
        StripString(line1, line1);
        if (line1[0] !='\0')
        {
          // create a new planet name object
          pPlanet = new CString;
          if (pPlanet == NULL)
          {
            // an error
            ErrorDialog.Report(ERR_OUTOFMEMORY, ERR_HEADER);
            return FALSE;
          }
          *pPlanet = line1;
          // add the planet
          pSystemPlanet->GetPlanets()->AddTail((CObject *)pPlanet);
        }
      } while (line[0] != '\0');

    }

  }
  // close the file
  AvailableSystemsPlanetsFile.Close();

  return TRUE;

}

//-----------------------------------
// find a system name
//-----------------------------------
BOOL CAvailableSystemsAndPlanets::FindSystemName(LONG Index, CString *SystemName)
{
  POSITION          pos;
  CSystemPlanetData *pSystemPlanet;

  // search the index
  pos = AvailableSystemsPlanetsList.FindIndex(Index);
  if (pos == NULL) return FALSE;

  // get the system name
  pSystemPlanet = (CSystemPlanetData *)AvailableSystemsPlanetsList.GetAt(pos);
  *SystemName = pSystemPlanet->GetSystemName();

  return TRUE;
}

//-----------------------------------
// find a system
//-----------------------------------
CSystemPlanetData *CAvailableSystemsAndPlanets::FindSystem(CString SystemName)
{

  // do it
  return AvailableSystemsPlanetsList.FindSystem(SystemName);
}

//-----------------------------------
// find a planet
//-----------------------------------
BOOL CAvailableSystemsAndPlanets::FindPlanet(CSystemPlanetData *pSystemPlanet, CString PlanetName)
{
  POSITION          pos;
  CString           *pPlanet;

  // loop through all planets
  for (pos = pSystemPlanet->GetPlanets()->GetHeadPosition(); pos != NULL; )
  {
    // get the planet name
    pPlanet = (CString *)pSystemPlanet->GetPlanets()->GetNext(pos);
    // compare it
    if (pPlanet->CompareNoCase(PlanetName) == 0) return TRUE;
  }

  return FALSE;
}

//-----------------------------------
// fill a listbox with the system names
//-----------------------------------
void CAvailableSystemsAndPlanets::FillListBoxWithSystems(CModalDialog *pDialog, LONG ListBoxId)
{
  POSITION          pos;
  CSystemPlanetData *pSystemPlanet;
  CString           CStr;

  // copy the system list to the list box
  for (pos = AvailableSystemsPlanetsList.GetHeadPosition(); pos != NULL; )
  {
    // get planet
    pSystemPlanet = (CSystemPlanetData *)AvailableSystemsPlanetsList.GetNext(pos);

    // put it in the box
    CStr = pSystemPlanet->GetSystemName();
    pDialog->SendDlgItemMessage(ListBoxId, LB_ADDSTRING,0,(LONG)(const char *)CStr);
  }
}


//-----------------------------------
// fill a listbox with the planet names of a given system
//-----------------------------------
void CAvailableSystemsAndPlanets::FillListBoxWithSystemPlanets(CSystemData *pSystem, CModalDialog *pDialog, LONG ListBoxId)
{
  POSITION          pos;
  CSystemPlanetData *pSystemPlanet;
  CString           CStr, *Planet;

  // first search the system
  CStr = pSystem->GetSystemName();
  pSystemPlanet = AvailableSystemsPlanetsList.FindSystem(CStr);
  if (pSystemPlanet != NULL)
  {
 
    // copy the planet list to the list box
    for (pos = pSystemPlanet->GetPlanets()->GetHeadPosition(); pos != NULL; )
    {
      // get the planet name
      Planet = (CString *)pSystemPlanet->GetPlanets()->GetNext(pos);
      // put it in the box
      pDialog->SendDlgItemMessage(ListBoxId, LB_ADDSTRING,0,(LONG)(const char *) *Planet);
    }
  }
}

